import numpy as np
try:
    import cPickle as pickle
except ImportError:
    import pickle
from abc import ABCMeta, abstractmethod
from six import add_metaclass

from glycresoft import symbolic_expression
from glycopeptidepy import HashableGlycanComposition
from glypy.structure.glycan_composition import FrozenMonosaccharideResidue as FMR


epsilon = 1e-6


def symbolic_composition(obj):
    """Extract the glycan composition from *obj* and converts it
    into an instance of :class:`.GlycanSymbolContext`

    The extraction process first attempts to access *obj.glycan_composition*
    but on an :class:`AttributeError` calls :meth:`.HashableGlycanComposition.parse`
    on *obj*. The resulting :class:`.GlycanComposition` object is then passed
    to :class:`.GlycanSymbolContext`

    Parameters
    ----------
    obj : object
        The object to convert

    Returns
    -------
    GlycanSymbolContext
    """
    try:
        composition = obj.glycan_composition
    except AttributeError:
        composition = HashableGlycanComposition.parse(obj)
    composition = symbolic_expression.GlycanSymbolContext(composition)
    return composition


class ScoringFeatureBase(object):
    """A base class for a type that either on instantiation
    or on invokation of :meth:`score` will produce a confidence
    metric for a (possibly annotated) chromatogram
    """
    name = None
    feature_type = None

    @classmethod
    def get_feature_type(self):
        return self.feature_type

    @classmethod
    def get_feature_name(self):
        name = getattr(self, "name", None)
        if name is None:
            return "%s:%s" % (self.get_feature_type(), self.__name__)
        else:
            return name

    @classmethod
    def reject(self, score_components, solution):
        score = score_components[self.get_feature_type()]
        return score < 0.15

    @classmethod
    def configure(cls, analysis_data):
        return {}


class DummyFeature(ScoringFeatureBase):
    def __init__(self, name, feature_type):
        self.name = name
        self.feature_type = feature_type

    def score(self, chromatogram, *args, **kwargs):
        return 0.985

    def get_feature_type(self):
        return self.feature_type

    def get_feature_name(self):
        name = getattr(self, "name", None)
        if name is None:
            name = self.__name__
        return "%s:%s" % (self.get_feature_type(), name)

    def reject(self, score_components, solution):
        score = score_components[self.feature_type]
        return score < 0.15

    def clone(self):
        return self


@add_metaclass(ABCMeta)
class DiscreteCountScoringModelBase(object):
    def __init__(self, *args, **kwargs):
        pass

    def score(self, chromatogram, *args, **kwargs):
        return 0

    def dump(self, file_obj):
        pickle.dump(self, file_obj)

    @classmethod
    def load(cls, file_obj):
        return pickle.load(file_obj)

    @abstractmethod
    def get_state_count(self, chromatogram):
        # Returns the number of states
        raise NotImplementedError()

    @abstractmethod
    def get_states(self, chromatogram):
        # Returns a list enumerating the states
        # this entity was observed in
        raise NotImplementedError()

    @abstractmethod
    def get_signal_proportions(self, chromatogram):
        # Returns a mapping from state to proportion of
        # total signal observed in state
        raise NotImplementedError()

    def reject(self, score_components, solution):
        score = score_components[self.feature_type]
        return score < 0.15


class UniformCountScoringModelBase(DiscreteCountScoringModelBase):
    def score(self, chromatogram, *args, **kwargs):
        return min(0.4 * self.get_state_count(chromatogram), 1.0) - epsilon


def decay(x, step=0.4, rate=1.5):
    v = 0
    for i in range(x):
        v += (step / (i + rate))
    return v


class DecayRateCountScoringModelBase(DiscreteCountScoringModelBase):
    def __init__(self, step=0.4, rate=1.5):
        self.step = step
        self.rate = rate

    def score(self, chromatogram, *args, **kwargs):
        k = self.get_state_count(chromatogram)
        return decay(k, self.step, self.rate) - epsilon

    def clone(self):
        return self.__class__(self.step, self.rate)


class LogarithmicCountScoringModelBase(DiscreteCountScoringModelBase):
    def __init__(self, steps=5):
        self.base = steps

    def _logarithmic_change_of_base(self, k):
        if k >= self.base:
            return 1.0
        return np.log(k) / np.log(self.base)

    def score(self, chromatogram, *args, **kwargs):
        # Ensure k > 1 so that the value is greater than 0.0
        # as `log(1) = 0`
        k = self.get_state_count(chromatogram) + 1
        return self._logarithmic_change_of_base(k) - epsilon

    def clone(self):
        return self.__class__(self.base)


def ones(x):
    return (x - (np.floor(x / 10.) * 10))


def neighborhood_of(x, scale=100.):
    n = x / scale
    up = ones(n) > 5
    if up:
        neighborhood = (np.floor(n / 10.) + 1) * 10
    else:
        neighborhood = (np.floor(n / 10.) + 1) * 10
    return neighborhood * scale


class MassScalingCountScoringModel(DiscreteCountScoringModelBase):
    def __init__(self, table, neighborhood_width=100.):
        self.table = table
        self.neighborhood_width = neighborhood_width

    def neighborhood_of(self, mass):
        n = mass / self.neighborhood_width
        up = ones(n) > 5
        if up:
            neighborhood = (np.floor(n / 10.) + 1) * 10
        else:
            neighborhood = (np.floor(n / 10.) + 1) * 10
        return neighborhood * self.neighborhood_width

    def get_neighborhood_key(self, neutral_mass):
        neighborhood = self.neighborhood_of(neutral_mass)
        return neighborhood

    def handle_missing_neighborhood(self, chromatogram, neighborhood, *args, **kwargs):
        return 0

    def handle_missing_bin(self, chromatogram, bins, key, neighborhood, *args, **kwargs):
        return sum(bins.values()) / float(len(bins))

    def transform_state(self, state):
        return state

    def score(self, chromatogram, *args, **kwargs):
        total = 0.
        neighborhood = self.get_neighborhood_key(chromatogram.neutral_mass)
        if neighborhood not in self.table:
            return self.handle_missing_neighborhood(
                chromatogram, neighborhood, *args, **kwargs)

        bins = self.table[neighborhood]

        for state in self.get_states(chromatogram):
            state = self.transform_state(state)
            try:
                total += bins[state]
            except KeyError:
                total += self.handle_missing_bin(
                    chromatogram, bins, state, neighborhood, *args, **kwargs)
        total = max(min(total, 1.0) - epsilon, epsilon)
        return total

    def clone(self):
        return self.__class__(self.table, self.neighborhood_width)


class ProportionBasedMassScalingCountScoringModel(MassScalingCountScoringModel):

    def kullback_leibler_divergence(self, observed, expected):
        total = 0
        for key in expected:
            try:
                observed_frequency = observed[key]
            except KeyError:
                observed_frequency = 1e-3
            total += observed_frequency * (np.log(observed_frequency) - np.log(expected[key]))
        return total

    def expnorm_kl(self, observed, expected):
        return 1 - np.exp(-self.kullback_leibler_divergence(observed, expected))

    def score(self, chromatogram, *args, **kwargs):
        neighborhood = self.get_neighborhood_key(chromatogram.neutral_mass)
        if neighborhood not in self.table:
            return self.handle_missing_neighborhood(
                chromatogram, neighborhood, *args, **kwargs)

        model_bin = self.table[neighborhood]
        observed_dist = dict()

        for key, proportion in self.get_signal_proportions(chromatogram):
            key = self.transform_state(key)
            observed_dist[key] = proportion
        score = self.expnorm_kl(observed_dist, model_bin)
        score = max(min(score, 1.0) - epsilon, epsilon)
        return score


class CompositionDispatchingModel(ScoringFeatureBase):
    def __init__(self, rule_model_map, default_model):
        self.rule_model_map = rule_model_map
        self.default_model = default_model

    def get_composition(self, obj):
        if obj.composition is not None:
            if obj.glycan_composition is not None:
                return symbolic_composition(obj)
        return None

    def find_model(self, composition):
        if composition is None:
            return self.default_model
        for rule in self.rule_model_map:
            if rule(composition):
                return self.rule_model_map[rule]

        return self.default_model

    def score(self, chromatogram, *args, **kwargs):
        composition = self.get_composition(chromatogram)
        model = self.find_model(composition)
        return model.score(chromatogram, *args, **kwargs)

    def get_feature_type(self):
        return self.default_model.get_feature_type()

    def get_feature_name(self):
        return "%s:%s(%s)" % (
            self.get_feature_type(), self.__class__.__name__,
            ", ".join([v.get_feature_name()
                       for v in self.rule_model_map.values()]))

    def reject(self, score_components, solution):
        # score = score_components[self.get_feature_type()]
        # return score < 0.15
        composition = self.get_composition(solution)
        model = self.find_model(composition)
        return model.reject(score_components, solution)

    def clone(self):
        return self.__class__(self.rule_model_map, self.default_model)


neuac = FMR.from_iupac_lite("NeuAc")
neugc = FMR.from_iupac_lite("NeuGc")
neu = FMR.from_iupac_lite("Neu")
sulfate = FMR.from_iupac_lite("@sulfate")
hexuronic_acid = FMR.from_iupac_lite("aHex")


def is_sialylated(composition):
    return (composition[neuac] + composition[neugc] + composition[neu]) > 0


def degree_of_sialylation(composition):
    return (composition[neuac] + composition[neugc] + composition[neu])


def is_high_mannose(composition):
    return (composition['HexNAc'] == 2 and composition['Hex'] > 3 and
            not is_sialylated(composition))



import time
import json
import logging

from collections import defaultdict, deque, namedtuple

import multiprocessing
from multiprocessing.pool import ThreadPool
from multiprocessing import Process, Event, JoinableQueue
from multiprocessing.managers import RemoteError

from queue import Empty as QueueEmptyException
from threading import RLock, Condition, Thread

import numpy as np

from glycresoft import serialize
from glycresoft.task import TaskBase, IPCLoggingManager

from glycresoft.database import (
    GlycanCompositionDiskBackedStructureDatabase, GlycopeptideDiskBackedStructureDatabase)


from glycresoft.database.composition_network import NeighborhoodWalker, make_n_glycan_neighborhoods
from glycresoft.composition_distribution_model import (
    smooth_network, display_table, VariableObservationAggregation,
    GlycanCompositionSolutionRecord, GlycomeModel)
from glycresoft.models import GeneralScorer, get_feature


from .glycosite_model import GlycanPriorRecord, GlycosylationSiteModel
from .glycoprotein_model import ProteinStub

_default_chromatogram_scorer = GeneralScorer.clone()
_default_chromatogram_scorer.add_feature(get_feature("null_charge"))

logger = logging.getLogger("glycresoft.glycosite_model")
logger.addHandler(logging.NullHandler())


def _truncate_name(name, limit=30):
    if len(name) > limit:
        return name[:limit - 3] + '...'
    return name



class FakeLock:
    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass


EmptySite = namedtuple("EmptySite", ("position", "protein_name"))


class GlycosylationSiteModelBuilder(TaskBase):
    _timeout_per_unit = 300

    def __init__(self, glycan_graph, chromatogram_scorer=None, belongingness_matrix=None,
                 unobserved_penalty_scale=None, lambda_limit=0.2,
                 require_multiple_observations=True,
                 observation_aggregator=None, n_threads=1):
        if observation_aggregator is None:
            observation_aggregator = VariableObservationAggregation
        if chromatogram_scorer is None:
            chromatogram_scorer = _default_chromatogram_scorer
        if unobserved_penalty_scale is None:
            unobserved_penalty_scale = 1.0

        self.network = glycan_graph
        if not self.network.neighborhoods:
            self.network.neighborhoods = make_n_glycan_neighborhoods()

        self.chromatogram_scorer = chromatogram_scorer
        self.belongingness_matrix = belongingness_matrix
        self.observation_aggregator = observation_aggregator
        self.require_multiple_observations = require_multiple_observations
        self.unobserved_penalty_scale = unobserved_penalty_scale
        self.lambda_limit = lambda_limit

        if self.belongingness_matrix is None:
            self.belongingness_matrix = self.build_belongingness_matrix()

        self.site_models = []
        self.n_threads = n_threads
        self._lock = RLock() if self.n_threads > 1 else FakeLock()
        self._concurrent_jobs = 0
        if self.n_threads > 1:
            self.thread_pool = ThreadPool(n_threads)
        else:
            self.thread_pool = None

    def build_belongingness_matrix(self):
        network = self.network
        neighborhood_walker = NeighborhoodWalker(
            network, network.neighborhoods)
        belongingness_matrix = neighborhood_walker.build_belongingness_matrix()
        return belongingness_matrix

    def _transform_glycopeptide(self, glycopeptide, evaluate_chromatograms=False):
        gp = glycopeptide
        if evaluate_chromatograms:
            ms1_score = self.chromatogram_scorer.logitscore(gp.chromatogram)
        else:
            ms1_score = gp.ms1_score
        return GlycanCompositionSolutionRecord(gp.glycan_composition, ms1_score, gp.total_signal)

    def prepare_glycoprotein(self, glycoprotein, evaluate_chromatograms=False):
        prepared = []
        for i, site in enumerate(glycoprotein.site_map['N-Linked'].sites):
            gps_for_site = glycoprotein.site_map[
                'N-Linked'][glycoprotein.site_map['N-Linked'].sites[i]]
            gps_for_site = [
                gp for gp in gps_for_site if gp.chromatogram is not None]

            self.log('... %d Identified Glycopeptides At Site %d for %s' %
                     (len(gps_for_site), site, _truncate_name(glycoprotein.name, )))

            glycopeptides = [
                gp for gp in gps_for_site if gp.chromatogram is not None]
            records = []
            for gp in glycopeptides:
                records.append(self._transform_glycopeptide(
                    gp, evaluate_chromatograms))
            prepared.append((records, site, ProteinStub(glycoprotein.name)))
        return prepared

    def add_glycoprotein(self, glycoprotein, evaluate_chromatograms=False):
        async_results = []
        sites_to_log = []
        for i, site in enumerate(glycoprotein.site_map['N-Linked'].sites):
            gps_for_site = glycoprotein.site_map[
                'N-Linked'][glycoprotein.site_map['N-Linked'].sites[i]]
            gps_for_site = [
                gp for gp in gps_for_site if gp.chromatogram is not None]

            self.log('... %d Identified Glycopeptides At Site %d for %s' %
                     (len(gps_for_site), site, _truncate_name(glycoprotein.name, )))

            glycopeptides = [
                gp for gp in gps_for_site if gp.chromatogram is not None]
            records = []
            for gp in glycopeptides:
                records.append(self._transform_glycopeptide(
                    gp, evaluate_chromatograms))

            if self.n_threads == 1:
                self.fit_site_model(records, site, glycoprotein)
            else:
                sites_to_log.append(site)
                with self._lock:
                    self._concurrent_jobs += 1
                    async_results.append(
                        self.thread_pool.apply_async(self.fit_site_model, (records, site, glycoprotein, )))
        if async_results:
            time.sleep(20)
            for i, result in enumerate(async_results):
                if not result.ready():
                    self.log("... Waiting for Result from Site %d of %s" % (
                        sites_to_log[i], _truncate_name(glycoprotein.name, )))
                result.get(self._timeout_per_unit * self.n_threads + 20)
                with self._lock:
                    self._concurrent_jobs -= 1
            with self._lock:
                self.log("... Finished Fitting %s, %d Tasks Pending" % (
                    _truncate_name(glycoprotein.name), self._concurrent_jobs))

    def _get_learnable_cases(self, observations):
        learnable_cases = [rec for rec in observations if rec.score > 1]
        if not learnable_cases:
            return []
        if self.require_multiple_observations:
            agg = VariableObservationAggregation(self.network)
            agg.collect(learnable_cases)
            recs, var = agg.build_records()
            # TODO: Rewrite to avoid using VariableObservationAggregation because calculation
            #       of the variance matrix is expensive.
            #
            # Use VariableObservationAggregation algorithm to collect the glycan
            # composition observations according to the network definition of multiple
            # observations, and then extract the observed indices along the diagonal
            # of the variance matrix.
            #
            # Those indices which are equal to 1.0 are those
            # where the glycan composition was only observed once, according to the
            # transformation VariableObservationAggregation carries out when estimating
            # the variance of each glycan composition observed.
            rec_variance = np.diag(var.variance_matrix)[var.observed_indices]
            stable_cases = set(
                [gc.glycan_composition for gc, v in zip(recs, rec_variance) if v != 1.0])
            self.log("... %d Stable Glycan Compositions" % (
                len(stable_cases)))
            if len(stable_cases) == 0:
                stable_cases = set([gc.glycan_composition for gc in recs])
                self.log("... No Stable Cases Found. Using %d Glycan Compositions" % (
                    len(stable_cases), ))
            if len(stable_cases) == 0:
                return []
        else:
            stable_cases = {
                case.glycan_composition for case in learnable_cases}
        learnable_cases = [
            rec for rec in learnable_cases
            if rec.score > 1 and rec.glycan_composition in stable_cases
        ]
        return learnable_cases

    def fit_site_model(self, observations, site, glycoprotein):
        learnable_cases = self._get_learnable_cases(observations)

        if not learnable_cases:
            return None

        acc = defaultdict(list)
        for case in learnable_cases:
            acc[case.glycan_composition].append(case)
        log_buffer = []
        log_buffer.append("... %d Glycan Compositions for Site %d of %s" % (
            len(acc), site, _truncate_name(glycoprotein.name, )))
        # for key, value in sorted(acc.items(), key=lambda x: x[0].mass()):
        #     log_buffer.append("... %s: [%s]" % (
        #         key,
        #         ', '.join(["%0.2f" % f for f in sorted(
        #             [r.score for r in value])])
        #     ))
        # self.log('\n'.join(log_buffer))

        fitted_network, search_result, params = smooth_network(
            self.network, learnable_cases,
            belongingness_matrix=self.belongingness_matrix,
            observation_aggregator=VariableObservationAggregation,
            annotate_network=False)
        if params is None:
            self.log("Skipping Site %d of %s" %
                     (site, _truncate_name(glycoprotein.name)))
            return None
        self.log("... Site %d of %s Lambda: %f" %
                 (site, _truncate_name(glycoprotein.name), params.lmbda,))
        display_table([x.name for x in self.network.neighborhoods],
                        np.array(params.tau).reshape((-1, 1)))
        updated_params = params.clone()
        updated_params.lmbda = min(self.lambda_limit, params.lmbda)
        self.log("... Projecting Solution Onto Network for Site %d of %s" %
                 (site, _truncate_name(glycoprotein.name)))
        fitted_network = search_result.annotate_network(updated_params)
        for node in fitted_network:
            if node.marked:
                node.score *= self.unobserved_penalty_scale

        site_distribution = dict(
            zip([x.name for x in self.network.neighborhoods], updated_params.tau.tolist()))
        glycan_map = {
            str(node.glycan_composition): GlycanPriorRecord(node.score, not node.marked)
            for node in fitted_network
        }
        site_model = GlycosylationSiteModel(
            glycoprotein.name,
            site,
            site_distribution,
            updated_params.lmbda,
            glycan_map)
        self.site_models.append(site_model.pack())
        return site_model

    def save_models(self, path):
        with open(path, 'wt') as fh:
            prepared = []
            for site in sorted(self.site_models, key=lambda x: (x.protein_name, x.position)):
                prepared.append(site.to_dict())
            json.dump(prepared, fh)


class GlycositeModelBuildingProcess(Process):
    process_name = "glycosylation-site-modeler"

    def __init__(self, builder, input_queue, output_queue, producer_done_event, output_done_event, log_handler):
        Process.__init__(self)
        self.builder = builder
        self.input_queue = input_queue
        self.output_queue = output_queue
        self.producer_done_event = producer_done_event
        self.output_done_event = output_done_event
        self.log_handler = log_handler

    def all_work_done(self):
        """A helper method to encapsulate :attr:`output_done_event`'s ``is_set``
        method.

        Returns
        -------
        bool
        """
        return self.output_done_event.is_set()

    def log(self, message):
        """Send a normal logging message via :attr:`log_handler`

        Parameters
        ----------
        message : str
            The message to log
        """
        logger.info(message)

    def debug(self, message):
        """Send a debugging message via :attr:`log_handler`

        Parameters
        ----------
        message : str
            The message to log
        """
        logger.debug(message)

    def handle_item(self, observations, site, glycoprotein):
        model = self.builder.fit_site_model(observations, site, glycoprotein)
        if model is None:
            model = EmptySite(site, glycoprotein.name)
        self.output_queue.put(model)
        # Do not accumulate models within the process
        self.builder.site_models = []

    def task(self):
        """The worker process's main loop where it will poll for new work items,
        process incoming work items and send them back to the master process.
        """
        has_work = True
        self.items_handled = 0
        strikes = 0
        self.log_handler.add_handler()
        while has_work:
            try:
                observations, site, glycoprotein = self.input_queue.get(True, 5)
                self.input_queue.task_done()
                strikes = 0
            except QueueEmptyException:
                if self.producer_done_event.is_set():
                    has_work = False
                    break
                else:
                    strikes += 1
                    if strikes % 1000 == 0:
                        self.log("... %d iterations without work for %r" %
                                 (strikes, self))
                    continue
            self.items_handled += 1
            try:
                self.handle_item(observations, site, glycoprotein)
            except Exception:
                import traceback
                message = "An error occurred while processing %r of %r on %r:\n%s" % (
                    site, glycoprotein.name, self, traceback.format_exc())
                self.log(message)
                break
        self.cleanup()

    def cleanup(self):
        self.output_done_event.set()

    def run(self):
        new_name = getattr(self, 'process_name', None)
        if new_name is not None:
            TaskBase().try_set_process_name(new_name)
        # The task might actually use the same logger from different threads
        # which causes a deadlock. This "fixes" that. by writing directly to STDOUT
        # at the cost of not being able to write to file instead.
        # TaskBase.log_to_stdout()
        self.output_done_event.clear()
        try:
            self.task()
        except Exception:
            import traceback
            self.log("An exception occurred while executing %r.\n%s" % (
                self, traceback.format_exc()))
            self.cleanup()


class GlycoproteinSiteModelBuildingWorkflowBase(TaskBase):
    def __init__(self, analyses, glycopeptide_database, glycan_database,
                 unobserved_penalty_scale=None, lambda_limit=0.2,
                 require_multiple_observations=True, observation_aggregator=None,
                 output_path=None, n_threads=None, q_value_threshold=0.05,
                 network=None, include_decoy_glycans=True):
        if observation_aggregator is None:
            observation_aggregator = VariableObservationAggregation
        if unobserved_penalty_scale is None:
            unobserved_penalty_scale = 1.0

        self.q_value_threshold = q_value_threshold
        self.analyses = analyses
        self.glycopeptide_database = glycopeptide_database
        self.glycan_database = glycan_database

        self.unobserved_penalty_scale = unobserved_penalty_scale
        self.lambda_limit = lambda_limit

        self.require_multiple_observations = require_multiple_observations
        self.observation_aggregator = observation_aggregator

        self.output_path = output_path
        self.network = network
        self.include_decoy_glycans = include_decoy_glycans

    @classmethod
    def from_paths(cls, analysis_paths_and_ids, glycopeptide_hypothesis_path, glycopeptide_hypothesis_id,
                   glycan_hypothesis_path, glycan_hypothesis_id, unobserved_penalty_scale=None,
                   lambda_limit=0.2, require_multiple_observations=True, observation_aggregator=None,
                   output_path=None, n_threads=4, q_value_threshold=0.05, network=None, include_decoy_glycans=True):
        gp_db = GlycopeptideDiskBackedStructureDatabase(
            glycopeptide_hypothesis_path, glycopeptide_hypothesis_id)
        gc_db = GlycanCompositionDiskBackedStructureDatabase(
            glycan_hypothesis_path, glycan_hypothesis_id)

        analyses = [serialize.AnalysisDeserializer(conn, analysis_id=an_id)
                    for conn, an_id in analysis_paths_and_ids]
        inst = cls(
            analyses, gp_db, gc_db, unobserved_penalty_scale=unobserved_penalty_scale,
            lambda_limit=lambda_limit, require_multiple_observations=require_multiple_observations,
            observation_aggregator=observation_aggregator, output_path=output_path,
            n_threads=n_threads, q_value_threshold=q_value_threshold, network=network,
            include_decoy_glycans=include_decoy_glycans
        )
        return inst

    def count_glycosites(self, glycoproteins):
        n_sites = sum(len(gp.site_map['N-Linked']) for gp in glycoproteins)
        return n_sites

    def make_glycan_network(self):
        if self.network is None:
            network = self.glycan_database.glycan_composition_network
        else:
            network = self.network
        if self.include_decoy_glycans:
            network = network.augment_with_decoys()
        network.create_edges()

        model = GlycomeModel([], network)
        belongingness_matrix = model.belongingness_matrix
        network.neighborhoods = model.neighborhood_walker.neighborhoods
        return network, belongingness_matrix

    def load_identified_glycoproteins_from_analysis(self, analysis):
        if not isinstance(analysis, serialize.Analysis):
            analysis = analysis.analysis
        idgps = analysis.aggregate_identified_glycoproteins(
            analysis.identified_glycopeptides.filter(
                serialize.IdentifiedGlycopeptide.q_value <= self.q_value_threshold))
        return idgps

    def build_reference_protein_map(self):
        proteins = list(self.glycopeptide_database.proteins)
        index = {p.name: p for p in proteins}
        return index

    def aggregate_identified_glycoproteins(self):
        acc = defaultdict(list)
        n = len(self.analyses)
        for i, analysis in enumerate(self.analyses, 1):
            self.log("... Loading Glycopeptides for %s (%d/%d)" %
                     (analysis.name, i, n))
            for idgp in self.load_identified_glycoproteins_from_analysis(analysis):
                acc[idgp.name].append(idgp)
        result = []
        self.log("Merging Glycoproteins Across Replicates")
        n = float(len(acc))
        i = 1.0
        last = 0.1
        should_log = False
        for name, duplicates in acc.items():
            if i / n > last:
                should_log = True
                last += 0.1
            if i % 100 == 0:
                should_log = True
            if should_log:
                self.log("... Merging %s (%d/%d, %0.2f%%)" %
                         (name, i, n, i * 100.0 / n))
                should_log = False
            agg = duplicates[0]
            result.append(agg.merge(*duplicates[1:]))
            i += 1
        return result

    def _fit_glycoprotein_site_models(self, glycoproteins, builder):
        raise NotImplementedError()

    def _init_builder(self, network, belongingness_matrix):
        builder = GlycosylationSiteModelBuilder(
            network, belongingness_matrix=belongingness_matrix,
            unobserved_penalty_scale=self.unobserved_penalty_scale,
            lambda_limit=self.lambda_limit,
            require_multiple_observations=self.require_multiple_observations,
            observation_aggregator=self.observation_aggregator,
            n_threads=self.n_threads)
        return builder

    def run(self):
        self.log("Building Belongingness Matrix")
        network, belongingness_matrix = self.make_glycan_network()

        builder = self._init_builder(network, belongingness_matrix)

        self.log("Aggregating Glycoproteins")
        glycoproteins = self.aggregate_identified_glycoproteins()
        glycoproteins = sorted(
            glycoproteins,
            key=lambda x: len(x.identified_glycopeptides),
            reverse=True)

        self._fit_glycoprotein_site_models(glycoproteins, builder)

        self.log("Saving Models")
        if self.output_path is not None:
            builder.save_models(self.output_path)


class ThreadedGlycoproteinSiteModelBuildingWorkflow(GlycoproteinSiteModelBuildingWorkflowBase):
    _timeout_per_unit_site = 300

    def __init__(self, analyses, glycopeptide_database, glycan_database,
                 unobserved_penalty_scale=None, lambda_limit=0.2,
                 require_multiple_observations=True, observation_aggregator=None,
                 output_path=None, n_threads=4, q_value_threshold=0.05, network=None, include_decoy_glycans=True):
        super(ThreadedGlycoproteinSiteModelBuildingWorkflow, self).__init__(
            analyses, glycopeptide_database, glycan_database,
            unobserved_penalty_scale, lambda_limit,
            require_multiple_observations, observation_aggregator,
            output_path,  q_value_threshold=q_value_threshold, network=network,
            include_decoy_glycans=include_decoy_glycans
        )

        self.n_threads = n_threads
        self.thread_pool = ThreadPool(self.n_threads)
        self._lock = RLock()
        self._count_barrier = Condition()
        self._concurrent_jobs = 0

    def thread_pool_saturated(self, ratio=1.0):
        with self._lock:
            jobs = self._concurrent_jobs
        return (jobs / float(self.n_threads)) >= ratio

    def _add_glycoprotein(self, glycoprotein, builder, k_sites):
        # Acquire the condition lock, then wait until the thread pool is empty
        # enough to do some work, then release the condition lock and do the work
        self._count_barrier.acquire()
        while self.thread_pool_saturated():
            self._count_barrier.wait()
        self._count_barrier.release()

        with self._lock:
            self._concurrent_jobs += k_sites

        # This should block until all site model fitting finishes.
        builder.add_glycoprotein(glycoprotein)

        with self._lock:
            self._concurrent_jobs -= k_sites

        # Acquire the condition lock, wake up the next thread waiting, and release
        # the condition lock
        self._count_barrier.acquire()
        self._count_barrier.notify()
        self._count_barrier.release()

    def _fit_glycoprotein_site_models(self, glycoproteins, builder):
        n = len(glycoproteins)
        n_sites = self.count_glycosites(glycoproteins)
        k_sites_acc = 0
        self.log(
            "Analyzing %d glycoproteins with %d occupied N-glycosites" % (n, n_sites))
        result_collector = deque()
        for i, gp in enumerate(glycoproteins, 1):
            k_sites = len(gp.site_map["N-Linked"])
            k_sites_acc += k_sites
            self.log("Building Model for \"%s\" with %d occupied N-glycosites %d/%d (%0.2f%%, %0.2f%% sites)" % (
                _truncate_name(gp.name), k_sites, i, n, i * 100.0 / n, k_sites_acc * 100.0 / n_sites))
            if self.n_threads == 1:
                builder.add_glycoprotein(gp)
            else:
                with self._lock:
                    result = self.thread_pool.apply_async(
                        self._add_glycoprotein, (gp, builder, k_sites))
                    result_collector.append((result, gp, k_sites))
                    # sleep briefly to allow newly queued task to start
                    time.sleep(1)
                # If the thread pool is full, we'll stop enqueuing new jobs and wait for it to clear out
                if self.thread_pool_saturated():
                    while self.thread_pool_saturated():
                        running_result, running_gp, running_gp_k_sites = result_collector.popleft()
                        while running_result.ready() and result_collector:
                            # get will re-raise errors if they occurred.
                            running_result.get()
                            running_result, running_gp, running_gp_k_sites = result_collector.popleft()
                        if not running_result.ready():
                            self.log("... Awaiting %s with %d Sites" % (
                                _truncate_name(running_gp.name), running_gp_k_sites))
                            running_result.get(
                                self._timeout_per_unit_site * running_gp_k_sites * self.n_threads)

        # Now drain any pending tasks.
        while result_collector:
            running_result, running_gp, running_gp_k_sites = result_collector.popleft()
            while running_result.ready() and result_collector:
                # get will re-raise errors if they occurred.
                running_result.get()
                running_result, running_gp, running_gp_k_sites = result_collector.popleft()
            if not running_result.ready():
                self.log("... Awaiting %s with %d Sites" % (
                    _truncate_name(running_gp.name), running_gp_k_sites))
                running_result.get(
                    self._timeout_per_unit_site * running_gp_k_sites * self.n_threads)


class MultiprocessingGlycoproteinSiteModelBuildingWorkflow(GlycoproteinSiteModelBuildingWorkflowBase):

    def __init__(self, analyses, glycopeptide_database, glycan_database,
                 unobserved_penalty_scale=None, lambda_limit=0.2,
                 require_multiple_observations=True, observation_aggregator=None,
                 output_path=None, n_threads=4, q_value_threshold=0.05, network=None,
                 include_decoy_glycans=True):
        super(MultiprocessingGlycoproteinSiteModelBuildingWorkflow, self).__init__(
            analyses, glycopeptide_database, glycan_database,
            unobserved_penalty_scale, lambda_limit,
            require_multiple_observations, observation_aggregator,
            output_path, q_value_threshold=q_value_threshold, network=network,
            include_decoy_glycans=include_decoy_glycans
        )

        self.builder = None

        self.input_queue = JoinableQueue(1000)
        self.output_queue = JoinableQueue(1000)
        self.input_done_event = Event()
        self.n_threads = 1
        self.n_workers = n_threads
        self.workers = []
        self._has_remote_error = False
        self.ipc_manager = IPCLoggingManager()

    def prepare_glycoprotein_for_dispatch(self, glycoprotein, builder):
        prepared = builder.prepare_glycoprotein(glycoprotein)
        return prepared

    def feed_queue(self, glycoproteins, builder):
        n = len(glycoproteins)
        n_sites = self.count_glycosites(glycoproteins)
        self.log(
            "Analyzing %d glycoproteins with %d occupied N-glycosites" % (n, n_sites))
        i_site = 0
        for glycoprotein in glycoproteins:
            prepared = self.prepare_glycoprotein_for_dispatch(glycoprotein, builder)
            for work_item in prepared:
                i_site += 1
                self.input_queue.put(work_item)
                if i_site % 50 == 0 and i_site != 0:
                    self.input_queue.join()
        self.input_done_event.set()

    def _handle_local(self, glycoproteins, builder, seen):
        for glycoprotein in glycoproteins:
            prepared = self.prepare_glycoprotein_for_dispatch(
                glycoprotein, builder)
            for records, site, protein_stub in prepared:
                key = (protein_stub.name, site)
                if key in seen:
                    continue
                else:
                    seen[key] = -1
                    model = builder.fit_site_model(records, site, protein_stub)
                    if model is not None:
                        self.builder.site_models.append(model)

    def make_workers(self):
        for _i in range(self.n_workers):
            worker = GlycositeModelBuildingProcess(
                self.builder, self.input_queue, self.output_queue,
                producer_done_event=self.input_done_event,
                output_done_event=Event(),
                log_handler=self.ipc_manager.sender())
            self.workers.append(worker)
            worker.start()

    def clear_pool(self):
        for _i, worker in enumerate(self.workers):
            exitcode = worker.exitcode
            if exitcode != 0 and exitcode is not None:
                self.log("... Worker Process %r had exitcode %r" % (worker, exitcode))
            try:
                worker.join(1)
            except AttributeError:
                pass
            if worker.is_alive():
                self.debug("... Worker Process %r is still alive and incomplete" % (worker, ))
                worker.terminate()

    def all_workers_finished(self):
        """Check if all worker processes have finished.
        """
        worker_still_busy = False
        assert self.workers
        for worker in self.workers:
            try:
                is_done = worker.all_work_done()
                if not is_done:
                    worker_still_busy = True
                    break
            except (RemoteError, KeyError):
                worker_still_busy = True
                self._has_remote_error = True
                break
        return not worker_still_busy

    def _fit_glycoprotein_site_models(self, glycoproteins, builder):
        self.builder = builder
        feeder_thread = Thread(target=self.feed_queue, args=(glycoproteins, builder))
        feeder_thread.daemon = True
        feeder_thread.start()
        self.make_workers()
        n_sites = self.count_glycosites(glycoproteins)
        seen = dict()
        strikes = 0
        start_time = time.time()
        i = 0
        has_work = True
        while has_work:
            try:
                site_model = self.output_queue.get(True, 3)
                self.output_queue.task_done()
                key = (site_model.protein_name, site_model.position)
                seen[(key)] = i
                if key in seen:
                    self.debug(
                        "...... Duplicate Results For %s. First seen at %r, now again at %r" % (
                            key, seen[key], i, ))
                else:
                    seen[key] = i
                i += 1
                strikes = 0
                if i % 1 == 0:
                    self.log(
                        "...... Processed %d sites (%0.2f%%)" % (i, i * 100. / n_sites))
                if not isinstance(site_model, EmptySite):
                    self.builder.site_models.append(site_model)
            except QueueEmptyException:
                if len(seen) == n_sites:
                    has_work = False
                # do worker life cycle management here
                elif self.all_workers_finished():
                    if len(seen) == n_sites:
                        has_work = False
                    else:
                        strikes += 1
                        if strikes % 25 == 0:
                            self.log(
                                "...... %d cycles without output (%d/%d, %0.2f%% Done)" % (
                                    strikes, len(seen), n_sites, len(seen) * 100. / n_sites))
                            self.debug("...... Processes")
                            for worker in self.workers:
                                self.debug("......... %r" % (worker,))
                            self.debug("...... IPC Manager: %r" % (self.ipc_manager,))
                        if strikes > 150:
                            self.log("Too much time has elapsed waiting for final results, finishing locally.")
                            self._handle_local(glycoproteins, builder, seen)
                else:
                    strikes += 1
                    if strikes % 50 == 0:
                        self.log(
                            "...... %d cycles without output (%d/%d, %0.2f%% Done, %d children still alive)" % (
                                strikes, len(seen), n_sites, len(
                                    seen) * 100. / n_sites,
                                len(multiprocessing.active_children()) - 1))
                        try:
                            input_queue_size = self.input_queue.qsize()
                        except Exception:
                            input_queue_size = -1
                        is_feeder_done = self.input_done_event.is_set()
                        self.log("...... Input Queue Status: %r. Is Feeder Done? %r" % (
                            input_queue_size, is_feeder_done))
                    if strikes > 150:
                        self.log("Too much time has elapsed waiting for workers, finishing locally.")
                        self._handle_local(glycoproteins, builder, seen)
                continue
        self.clear_pool()
        self.ipc_manager.stop()
        feeder_thread.join()
        dispatcher_end = time.time()
        self.log("... Dispatcher Finished (%0.3g sec.)" %
                 (dispatcher_end - start_time))


GlycoproteinSiteModelBuildingWorkflow = MultiprocessingGlycoproteinSiteModelBuildingWorkflow



# Copyright (c) 2018, Joshua Klein
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dill as _dill
import warnings
from sqlalchemy import exc as sa_exc

try:
    warnings.simplefilter("ignore", category=sa_exc.SAWarning)
    warnings.filterwarnings(action="ignore", module="SPARQLWrapper")
    warnings.filterwarnings(
        action="ignore",
        category=DeprecationWarning,
        module="pysqlite2.dbapi2")
finally:
    pass

from glycresoft import serialize

from glycresoft.piped_deconvolve import (
    ScanGenerator
)

from glycresoft.chromatogram_tree import (
    MassShift, CompoundMassShift, Chromatogram, ChromatogramTreeList,
    ChromatogramTreeNode, ChromatogramInterface, ChromatogramFilter,
    mass_shift)


from glycresoft.trace import (
    ChromatogramExtractor, ChromatogramProcessor)

from glycresoft import database
from glycresoft.database import (
    NeutralMassDatabase, GlycopeptideDiskBackedStructureDatabase,
    GlycanCompositionDiskBackedStructureDatabase)

from glycresoft import profiler

from glycresoft.config.config_file import get_configuration

get_configuration()

__all__ = [
    "ScanGenerator",
    "MassShift", "CompoundMassShift", "Chromatogram",
    "ChromatogramTreeNode", "ChromatogramTreeList",
    "ChromatogramInterface", "ChromatogramFilter",
    "mass_shift", "ChromatogramExtractor", "ChromatogramProcessor",
    "NeutralMassDatabase", "GlycopeptideDiskBackedStructureDatabase",
    "GlycanCompositionDiskBackedStructureDatabase", "serialize",
    "profiler", "plotting"
]



from scipy.misc import comb
import numpy as np

import math
from collections import defaultdict

from glypy.utils import make_struct


#: Store whether a fragment was observed for the reducing or non-reducing end of a particular glycosidic bond
GlycosidicBondMatch = make_struct("GlycosidicBondMatch", ["reducing", "non_reducing"])

#: Store how many times a fragment was observed for the reducing, non-reducing, or non-bisecting cross-ring
#: permutations of a particular monosaccharide
CrossringMatch = make_struct("CrossringMatch", ["reducing", "non_reducing", "non_bisecting"])


def is_primary_fragment(f):
    return f.break_count == 1


class FragmentScorer(object):
    """Score the data with respect to structure characterization.

    Attributes
    ----------
    crossring_map : dict
        Mapping of residue id to CrossringMatch
    crossring_score : float
        Score derived from cross-ring coverage
    final_score : float
        The final product score combining glycosidic and cross-ring scores
    fragments : list
        The list of matched fragments
    glycosidic_map : dict
        Mapping of link id to GlycosidicBondMatch
    glycosidic_score : float
        Score derived from glycosidic fragments
    structure : StructureRecord
        Source structure to be matched
    """
    def __init__(self, structure, matched_fragments):
        self.structure = structure
        self.fragments = list(filter(is_primary_fragment, matched_fragments))

        self.glycosidic_map = defaultdict(lambda: GlycosidicBondMatch(0, 0))
        self.crossring_map = defaultdict(lambda: CrossringMatch(0, 0, 0))

        self.glycosidic_score = 0.
        self.crossring_score = 0.

        self._score_glycosidic()
        self._score_crossring()

        self.final_score = 0.
        self._score_final()

    def _score_final(self):
        self.final_score = -100 * np.log10((1 - self.glycosidic_score) * (1 - self.crossring_score))

    def _score_glycosidic(self):
        for f in [f for f in self.fragments if not f.crossring_cleavages]:
            link_ind = f.link_ids.keys()[0]
            link_rec = self.glycosidic_map[link_ind]
            if f.kind in "BC":
                link_rec.non_reducing = 1
            else:
                link_rec.reducing = 1

        self.glycosidic_score = sum(map(sum, self.glycosidic_map.values())) / (2. * (len(self.structure) - 1))

    def _score_crossring(self):

        def has_parent(fragment, residue):
            parents = {n.id for p, n in residue.parents()}
            return bool(parents & fragment.included_nodes)

        def has_child(fragment, residue):
            children = {n.id for p, n in residue.children()}
            return bool(children & fragment.included_nodes)

        for f in [f for f in self.fragments if not f.link_ids]:
            residue = self.structure.get(f.crossring_cleavages.keys()[0])
            residue_record = self.crossring_map[residue.id]

            parent = has_parent(f, residue)
            child = has_child(f, residue)

            if parent and child:
                residue_record.non_bisecting += 1
            elif parent:
                residue_record.reducing += 1
            elif child:
                residue_record.non_reducing += 1

        self.crossring_score = sum(map(sum, self.crossring_map.values())) / (3. * len(self.structure))


def binomial_tail_probability(n, k, p):
    total = 0.0
    for i in range(k, n):
        total += comb(n, i, exact=True) * (p ** i) * ((1 - p) ** (n - i))
    return total


def drop_below_intensity(peaks, threshold):
    return [p for p in peaks if p.intensity > threshold]


def median_intensity(peaks):
    return np.median([p.intensity for p in peaks])


SimplePeak = make_struct("SimplePeak", ["intensity"])


epsilon = 1e-10


class IntensityRankScorer(object):
    """Scores an assigned peak list against a uniform intensity
    random model.

    Attributes
    ----------
    assigned_peaks : list
        The collection of assigned and deconvoluted peaks
    final_score : float
        The final score for this data
    medians : dict
        A cache for already computed median intensities
    peaklist : list
        The set of all experimental peaks
    scores : dict
        Scores for each tier
    transformed_peaks : list
        The collection of assigned peaks' most abundant peaks
    """
    def __init__(self, peaklist, assigned_peaks):
        self.peaklist = peaklist
        self.assigned_peaks = assigned_peaks
        self.transformed_peaks = [SimplePeak(
            max([p.intensity for p in sol.fit.experimental])) for sol in assigned_peaks]
        self.scores = {}
        self.medians = {0: 0}
        self.final_score = 0.
        self.rank()

    def rank(self):
        for a in range(1, 5):
            med = median_intensity(
                drop_below_intensity(
                    self.transformed_peaks,
                    self.medians[a - 1]))
            self.medians[a] = med
            i = len(drop_below_intensity(self.transformed_peaks, med))
            n = len(drop_below_intensity(self.peaklist, self.medians[a - 1]))
            p = binomial_tail_probability(n, i, 0.5)
            self.scores[a] = p
        total = 0.
        for p in self.scores.values():
            total += math.log(p + epsilon)
        self.final_score = -np.log10(math.exp(total))



from glycresoft.serialize.hypothesis import GlycanHypothesis
from glycresoft.serialize.hypothesis.glycan import (
    GlycanComposition as DBGlycanComposition,
    GlycanCombination, GlycanCombinationGlycanComposition)
from glycresoft.serialize import (
    DatabaseBoundOperation, Analysis, GlycanCompositionChromatogram,
    IdentifiedGlycopeptide, Glycopeptide)

from .glycan_source import (
    GlycanHypothesisSerializerBase, formula, GlycanCompositionToClass)

from glypy import GlycanComposition


class GlycanAnalysisHypothesisSerializer(GlycanHypothesisSerializerBase):
    def __init__(self, input_connection, analysis_id, hypothesis_name, output_connection=None):
        if output_connection is None:
            output_connection = input_connection
        self.input_connection = DatabaseBoundOperation(input_connection)
        self.output_connection = DatabaseBoundOperation(output_connection)
        GlycanHypothesisSerializerBase.__init__(self, output_connection, hypothesis_name)
        self.analysis_id = analysis_id
        self.seen_cache = set()

    def extract_composition(self, glycan_chromatogram):
        composition = GlycanComposition.parse(glycan_chromatogram.glycan_composition.serialize())
        if str(composition) in self.seen_cache:
            return
        self.seen_cache.add(str(composition))
        mass = composition.mass()
        composition_string = composition.serialize()
        formula_string = formula(composition.total_composition())
        inst = DBGlycanComposition(
            calculated_mass=mass, formula=formula_string,
            composition=composition_string,
            hypothesis_id=self.hypothesis_id)

        self.output_connection.session.add(inst)
        self.output_connection.session.flush()
        db_obj = self.input_connection.query(DBGlycanComposition).get(glycan_chromatogram.composition.id)
        for sc in db_obj.structure_classes:
            self.output_connection.session.execute(
                GlycanCompositionToClass.insert(), dict(glycan_id=inst.id, class_id=sc.id))
        self.output_connection.session.flush()

    def run(self):
        q = self.input_connection.session.query(GlycanCompositionChromatogram).filter(
            GlycanCompositionChromatogram.analysis_id == self.analysis_id)
        for gc in q:
            self.extract_composition(gc)
        self.output_connection.session.commit()


class GlycopeptideAnalysisGlycanCompositionExtractionHypothesisSerializer(GlycanHypothesisSerializerBase):
    def __init__(self, input_connection, analysis_id, hypothesis_name, output_connection=None):
        if output_connection is None:
            output_connection = input_connection
        self.input_connection = DatabaseBoundOperation(input_connection)
        self.output_connection = DatabaseBoundOperation(output_connection)
        GlycanHypothesisSerializerBase.__init__(self, output_connection, hypothesis_name)
        self.analysis_id = analysis_id
        self.seen_cache = set()

    def get_all_compositions(self):
        return self.input_connection.query(DBGlycanComposition).join(GlycanCombinationGlycanComposition).join(
            Glycopeptide,
            GlycanCombinationGlycanComposition.c.combination_id == Glycopeptide.glycan_combination_id).join(
            IdentifiedGlycopeptide, IdentifiedGlycopeptide.structure_id == Glycopeptide.id).filter(
            IdentifiedGlycopeptide.analysis_id == self.analysis_id)

    def extract_composition(self, db_obj):
        composition = GlycanComposition.parse(db_obj.composition)
        if str(composition) in self.seen_cache:
            return
        self.seen_cache.add(str(composition))
        mass = composition.mass()
        composition_string = composition.serialize()
        formula_string = formula(composition.total_composition())
        inst = DBGlycanComposition(
            calculated_mass=mass, formula=formula_string,
            composition=composition_string,
            hypothesis_id=self.hypothesis_id)
        self.output_connection.session.add(inst)
        self.output_connection.session.flush()
        for sc in db_obj.structure_classes:
            self.output_connection.session.execute(
                GlycanCompositionToClass.insert(), dict(glycan_id=inst.id, class_id=sc.id))
        self.output_connection.session.flush()

    def run(self):
        q = self.get_all_compositions()
        for gc in q:
            self.extract_composition(gc)
        self.output_connection.session.commit()



from __future__ import print_function

import numpy as np
from scipy.linalg import lapack


def display_table(names, values, sigfig=3, filter_empty=1, print_fn=None):
    if print_fn is None:
        print_fn = print
    values = np.array(values)
    maxlen = len(max(names, key=len)) + 2
    fstring = ("%%0.%df" % sigfig)
    for i in range(len(values)):
        if values[i, :].sum() or not filter_empty:
            print_fn(names[i].ljust(maxlen) + ('|'.join([fstring % f for f in values[i, :]])))


# Cholesky decomposition-based inverse based upon
# https://stackoverflow.com/a/58719188/1137920

inds_cache = {}


def _upper_triangular_to_symmetric(ut: np.ndarray):
    n = ut.shape[0]
    try:
        inds = inds_cache[n]
    except KeyError:
        inds = np.tri(n, k=-1, dtype=bool)
        inds_cache[n] = inds
    ut[inds] = ut.T[inds]


def fast_positive_definite_inverse(M: np.ndarray) -> np.ndarray:
    if M.size == 0:
        return np.zeros_like(M)

    # Compute the Cholesky decomposition of a symmetric positive-definite matrix
    cholesky, info = lapack.dpotrf(M)
    if info != 0:
        raise ValueError('dpotrf failed on input {}'.format(M))
    # Compute the inverse of `M` from its Cholesky decomposition
    inv, info = lapack.dpotri(cholesky)
    if info != 0:
        raise ValueError('dpotri failed on input {}'.format(cholesky))
    _upper_triangular_to_symmetric(inv)
    return inv



import unittest
import tempfile
import warnings

from glycresoft.serialize.hypothesis.peptide import Peptide, Protein, Glycopeptide
from glycresoft.database.builder.glycopeptide import naive_glycopeptide
from glycresoft.database.builder.glycan import (
    TextFileGlycanHypothesisSerializer,
    CombinatorialGlycanHypothesisSerializer)
from glycresoft import serialize
from . import fixtures

from .test_constrained_combinatorics import FILE_SOURCE as GLYCAN_RULE_FILE_SOURCE

from glycopeptidepy.structure import modification


FASTA_FILE_SOURCE = """
>sp|P02763|A1AG1_HUMAN Alpha-1-acid glycoprotein 1 OS=Homo sapiens GN=ORM1 PE=1 SV=1
MALSWVLTVLSLLPLLEAQIPLCANLVPVPITNATLDQITGKWFYIASAFRNEEYNKSVQ
EIQATFFYFTPNKTEDTIFLREYQTRQDQCIYNTTYLNVQRENGTISRYVGGQEHFAHLL
ILRDTKTYMLAFDVNDEKNWGLSVYADKPETTKEQLGEFYEALDCLRIPKSDVVYTDWKK
DKCEPLEKQHEKERKQEEGES

>sp|P19652|A1AG2_HUMAN Alpha-1-acid glycoprotein 2 OS=Homo sapiens GN=ORM2 PE=1 SV=2
MALSWVLTVLSLLPLLEAQIPLCANLVPVPITNATLDRITGKWFYIASAFRNEEYNKSVQ
EIQATFFYFTPNKTEDTIFLREYQTRQNQCFYNSSYLNVQRENGTVSRYEGGREHVAHLL
FLRDTKTLMFGSYLDDEKNWGLSFYADKPETTKEQLGEFYEALDCLCIPRSDVMYTDWKK
DKCEPLEKQHEKERKQEEGES
"""

HEPARANASE = """
>sp|Q9Y251|HPSE_HUMAN Heparanase OS=Homo sapiens GN=HPSE PE=1 SV=2
MLLRSKPALPPPLMLLLLGPLGPLSPGALPRPAQAQDVVDLDFFT
QEPLHLVSPSFLSVTIDANLATDPRFLILLGSPKLRTLARGLSPA
YLRFGGTKTDFLIFDPKKESTFEERSYWQSQVNQDICKYGSIPPD
VEEKLRLEWPYQEQLLLREHYQKKFKNSTYSRSSVDVLYTFANCS
GLDLIFGLNALLRTADLQWNSSNAQLLLDYCSSKGYNISWELGNE
PNSFLKKADIFINGSQLGEDFIQLHKLLRKSTFKNAKLYGPDVGQ
PRRKTAKMLKSFLKAGGEVIDSVTWHHYYLNGRTATKEDFLNPDV
LDIFISSVQKVFQVVESTRPGKKVWLGETSSAYGGGAPLLSDTFA
AGFMWLDKLGLSARMGIEVVMRQVFFGAGNYHLVDENFDPLPDYW
LSLLFKKLVGTKVLMASVQGSKRRKLRVYLHCTNTDNPRYKEGDL
TLYAINLHNVTKYLRLPYPFSNKQVDKYLLRPLGPHGLLSKSVQL
NGLTLKMVDDQTLPPLMEKPLRPGSSLGLPAFSYSFFVIRNAKVA
ACI
"""

VERSICAN = """
>sp|P13611|CSPG2_HUMAN Versican core protein OS=Homo sapiens GN=VCAN PE=1 SV=3
MFINIKSILWMCSTLIVTHALHKVKVGKSPPVRGSLSGKVSLPCHFSTMPTLPPSYNTSEFLRIKWSKIEVDKNGKDLKETTVLVAQNGN
IKIGQDYKGRVSVPTHPEAVGDASLTVVKLLASDAGLYRCDVMYGIEDTQDTVSLTVDGVVFHYRAATSRYTLNFEAAQKACLDVGAVIA
TPEQLFAAYEDGFEQCDAGWLADQTVRYPIRAPRVGCYGDKMGKAGVRTYGFRSPQETYDVYCYVDHLDGDVFHLTVPSKFTFEEAAKEC
ENQDARLATVGELQAAWRNGFDQCDYGWLSDASVRHPVTVARAQCGGGLLGVRTLYRFENQTGFPPPDSRFDAYCFKPKEATTIDLSILA
ETASPSLSKEPQMVSDRTTPIIPLVDELPVIPTEFPPVGNIVSFEQKATVQPQAITDSLATKLPTPTGSTKKPWDMDDYSPSASGPLGKL
DISEIKEEVLQSTTGVSHYATDSWDGVVEDKQTQESVTQIEQIEVGPLVTSMEILKHIPSKEFPVTETPLVTARMILESKTEKKMVSTVS
ELVTTGHYGFTLGEEDDEDRTLTVGSDESTLIFDQIPEVITVSKTSEDTIHTHLEDLESVSASTTVSPLIMPDNNGSSMDDWEERQTSGR
ITEEFLGKYLSTTPFPSQHRTEIELFPYSGDKILVEGISTVIYPSLQTEMTHRRERTETLIPEMRTDTYTDEIQEEITKSPFMGKTEEEV
FSGMKLSTSLSEPIHVTESSVEMTKSFDFPTLITKLSAEPTEVRDMEEDFTATPGTTKYDENITTVLLAHGTLSVEAATVSKWSWDEDNT
TSKPLESTEPSASSKLPPALLTTVGMNGKDKDIPSFTEDGADEFTLIPDSTQKQLEEVTDEDIAAHGKFTIRFQPTTSTGIAEKSTLRDS
TTEEKVPPITSTEGQVYATMEGSALGEVEDVDLSKPVSTVPQFAHTSEVEGLAFVSYSSTQEPTTYVDSSHTIPLSVIPKTDWGVLVPSV
PSEDEVLGEPSQDILVIDQTRLEATISPETMRTTKITEGTTQEEFPWKEQTAEKPVPALSSTAWTPKEAVTPLDEQEGDGSAYTVSEDEL
LTGSERVPVLETTPVGKIDHSVSYPPGAVTEHKVKTDEVVTLTPRIGPKVSLSPGPEQKYETEGSSTTGFTSSLSPFSTHITQLMEETTT
EKTSLEDIDLGSGLFEKPKATELIEFSTIKVTVPSDITTAFSSVDRLHTTSAFKPSSAITKKPPLIDREPGEETTSDMVIIGESTSHVPP
TTLEDIVAKETETDIDREYFTTSSPPATQPTRPPTVEDKEAFGPQALSTPQPPASTKFHPDINVYIIEVRENKTGRMSDLSVIGHPIDSE
SKEDEPCSEETDPVHDLMAEILPEFPDIIEIDLYHSEENEEEEEECANATDVTTTPSVQYINGKHLVTTVPKDPEAAEARRGQFESVAPS
QNFSDSSESDTHPFVIAKTELSTAVQPNESTETTESLEVTWKPETYPETSEHFSGGEPDVFPTVPFHEEFESGTAKKGAESVTERDTEVG
HQAHEHTEPVSLFPEESSGEIAIDQESQKIAFARATEVTFGEEVEKSTSVTYTPTIVPSSASAYVSEEEAVTLIGNPWPDDLLSTKESWV
EATPRQVVELSGSSSIPITEGSGEAEEDEDTMFTMVTDLSQRNTTDTLITLDTSRIITESFFEVPATTIYPVSEQPSAKVVPTKFVSETD
TSEWISSTTVEEKKRKEEEGTTGTASTFEVYSSTQRSDQLILPFELESPNVATSSDSGTRKSFMSLTTPTQSEREMTDSTPVFTETNTLE
NLGAQTTEHSSIHQPGVQEGLTTLPRSPASVFMEQGSGEAAADPETTTVSSFSLNVEYAIQAEKEVAGTLSPHVETTFSTEPTGLVLSTV
MDRVVAENITQTSREIVISERLGEPNYGAEIRGFSTGFPLEEDFSGDFREYSTVSHPIAKEETVMMEGSGDAAFRDTQTSPSTVPTSVHI
SHISDSEGPSSTMVSTSAFPWEEFTSSAEGSGEQLVTVSSSVVPVLPSAVQKFSGTASSIIDEGLGEVGTVNEIDRRSTILPTAEVEGTK
APVEKEEVKVSGTVSTNFPQTIEPAKLWSRQEVNPVRQEIESETTSEEQIQEEKSFESPQNSPATEQTIFDSQTFTETELKTTDYSVLTT
KKTYSDDKEMKEEDTSLVNMSTPDPDANGLESYTTLPEATEKSHFFLATALVTESIPAEHVVTDSPIKKEESTKHFPKGMRPTIQESDTE
LLFSGLGSGEEVLPTLPTESVNFTEVEQINNTLYPHTSQVESTSSDKIEDFNRMENVAKEVGPLVSQTDIFEGSGSVTSTTLIEILSDTG
AEGPTVAPLPFSTDIGHPQNQTVRWAEEIQTSRPQTITEQDSNKNSSTAEINETTTSSTDFLARAYGFEMAKEFVTSAPKPSDLYYEPSG
EGSGEVDIVDSFHTSATTQATRQESSTTFVSDGSLEKHPEVPSAKAVTADGFPTVSVMLPLHSEQNKSSPDPTSTLSNTVSYERSTDGSF
QDRFREFEDSTLKPNRKKPTENIIIDLDKEDKDLILTITESTILEILPELTSDKNTIIDIDHTKPVYEDILGMQTDIDTEVPSEPHDSND
ESNDDSTQVQEIYEAAVNLSLTEETFEGSADVLASYTQATHDESMTYEDRSQLDHMGFHFTTGIPAPSTETELDVLLPTATSLPIPRKSA
TVIPEIEGIKAEAKALDDMFESSTLSDGQAIADQSEIIPTLGQFERTQEEYEDKKHAGPSFQPEFSSGAEEALVDHTPYLSIATTHLMDQ
SVTEVPDVMEGSNPPYYTDTTLAVSTFAKLSSQTPSSPLTIYSGSEASGHTEIPQPSALPGIDVGSSVMSPQDSFKEIHVNIEATFKPSS
EEYLHITEPPSLSPDTKLEPSEDDGKPELLEEMEASPTELIAVEGTEILQDFQNKTDGQVSGEAIKMFPTIKTPEAGTVITTADEIELEG
ATQWPHSTSASATYGVEAGVVPWLSPQTSERPTLSSSPEINPETQAALIRGQDSTIAASEQQVAARILDSNDQATVNPVEFNTEVATPPF
SLLETSNETDFLIGINEESVEGTAIYLPGPDRCKMNPCLNGGTCYPTETSYVCTCVPGYSGDQCELDFDECHSNPCRNGATCVDGFNTFR
CLCLPSYVGALCEQDTETCDYGWHKFQGQCYKYFAHRRTWDAAERECRLQGAHLTSILSHEEQMFVNRVGHDYQWIGLNDKMFEHDFRWT
DGSTLQYENWRPNQPDSFFSAGEDCVVIIWHENGQWNDVPCNYHLTYTCKKGTVACGQPPVVENAKTFGKMKPRYEINSLIRYHCKDGFI
QRHLPTIRCLGNGRWAIPKITCMNPSAYQRTYSMKYFKNSSSAKDNSINTSKHDHRWSRRWQESRR
"""

o_glycans = """
{Hex:1; HexNAc:1; Neu5Ac:2}    O-Glycan
"""

simple_n_glycans = """
{Hex:5; HexNAc:2}    N-Glycan
"""

decorin = """
>sp|P21793|PGS2_BOVIN Decorin
MKATIIFLLVAQVSWAGPFQQKGLFDFMLEDEASGIGPEEHFPEVPEIEPMGPVCPFRCQ
CHLRVVQCSDLGLEKVPKDLPPDTALLDLQNNKITEIKDGDFKNLKNLHTLILINNKISK
ISPGAFAPLVKLERLYLSKNQLKELPEKMPKTLQELRVHENEITKVRKSVFNGLNQMIVV
ELGTNPLKSSGIENGAFQGMKKLSYIRIADTNITTIPQGLPPSLTELHLDGNKITKVDAA
SLKGLNNLAKLGLSFNSISAVDNGSLANTPHLRELHLNNNKLVKVPGGLADHKYIQVVYL
HNNNISAIGSNDFCPPGYNTKKASYSGVSLFSNPVQYWEIQPSTFRCVYVRAAVQLGNYK
"""

constant_modifications = ["Carbamidomethyl (C)"]
variable_modifications = ["Deamidation (N)", "Pyro-glu from Q (Q@N-term)"]


mt = modification.RestrictedModificationTable(
    constant_modifications=constant_modifications,
    variable_modifications=variable_modifications)

variable_modifications = [mt[v] for v in variable_modifications]
constant_modifications = [mt[c] for c in constant_modifications]


class FastaGlycopeptideTests(unittest.TestCase):

    def setup_tempfile(self, source):
        file_name = tempfile.mktemp() + '.tmp'
        open(file_name, 'w').write(source)
        return file_name

    def clear_file(self, path):
        open(path, 'wb')

    def test_build_hypothesis(self):
        glycan_file = self.setup_tempfile(GLYCAN_RULE_FILE_SOURCE)
        fasta_file = self.setup_tempfile(FASTA_FILE_SOURCE)
        db_file = fasta_file + '.db'

        glycan_builder = CombinatorialGlycanHypothesisSerializer(glycan_file, db_file)
        glycan_builder.start()

        glycopeptide_builder = naive_glycopeptide.MultipleProcessFastaGlycopeptideHypothesisSerializer(
            fasta_file, db_file, glycan_builder.hypothesis_id,
            use_uniprot=True,
            constant_modifications=constant_modifications,
            variable_modifications=variable_modifications,
            max_missed_cleavages=1)
        glycopeptide_builder.start()

        without_uniprot = 201400
        with_uniprot_without_variable_signal_peptide = 231800
        with_uniprot = 353400

        observed_count = glycopeptide_builder.query(Glycopeptide).count()

        assert observed_count in (
            with_uniprot, with_uniprot_without_variable_signal_peptide, without_uniprot)

        if observed_count == without_uniprot:
            warnings.warn("UniProt Annotations Not Used")
        if observed_count == with_uniprot_without_variable_signal_peptide:
            warnings.warn("Variable Signal Peptide Was Not Used")

        redundancy = glycopeptide_builder.query(
            Glycopeptide.glycopeptide_sequence,
            Protein.name,
            serialize.func.count(Glycopeptide.glycopeptide_sequence)).join(
            Glycopeptide.protein).join(Glycopeptide.peptide).group_by(
                Glycopeptide.glycopeptide_sequence,
                Protein.name,
                Peptide.start_position,
                Peptide.end_position).yield_per(1000)

        for sequence, protein, count in redundancy:
            self.assertEqual(count, 1, "%s in %s has multiplicity %d" % (sequence, protein, count))

        for case in glycopeptide_builder.query(Glycopeptide).filter(
                Glycopeptide.glycopeptide_sequence ==
                "SVQEIQATFFYFTPN(N-Glycosylation)K{Hex:5; HexNAc:4; Neu5Ac:2}").all():
            self.assertAlmostEqual(case.calculated_mass, 4123.718954557139, 5)

        self.clear_file(glycan_file)
        self.clear_file(fasta_file)
        self.clear_file(db_file)

    def test_missing_glycopeptide(self):
        glycan_file = self.setup_tempfile(o_glycans)
        fasta_file = self.setup_tempfile(HEPARANASE)
        db_file = fasta_file + '.db'

        glycan_builder = TextFileGlycanHypothesisSerializer(glycan_file, db_file)
        glycan_builder.start()

        glycopeptide_builder = naive_glycopeptide.FastaGlycopeptideHypothesisSerializer(
            fasta_file, db_file, glycan_builder.hypothesis_id, protease='trypsin',
            constant_modifications=constant_modifications,
            max_missed_cleavages=2)
        glycopeptide_builder.start()

        case = glycopeptide_builder.query(
            Glycopeptide.glycopeptide_sequence == "KFKNSTYS(O-Glycosylation)R{Hex:1; HexNAc:1; Neu5Ac:2}").first()
        self.assertIsNotNone(case)

        self.clear_file(glycan_file)
        self.clear_file(fasta_file)
        self.clear_file(db_file)

    def test_throughput(self):
        fasta_file = fixtures.get_test_data("phil-82-proteins.fasta")
        glycan_file = fixtures.get_test_data("IAV_matched_glycans.txt")
        db_file = self.setup_tempfile("")
        print(db_file)

        glycan_builder = TextFileGlycanHypothesisSerializer(glycan_file, db_file)
        glycan_builder.start()

        glycopeptide_builder = naive_glycopeptide.MultipleProcessFastaGlycopeptideHypothesisSerializer(
            fasta_file, db_file, glycan_builder.hypothesis_id, constant_modifications=constant_modifications,
            variable_modifications=variable_modifications, max_missed_cleavages=2)
        glycopeptide_builder.start()
        self.clear_file(db_file)

    def test_uniprot_info(self):
        fasta_file = self.setup_tempfile(decorin)
        glycan_file = self.setup_tempfile(o_glycans)
        db_file = self.setup_tempfile("")
        glycan_builder = TextFileGlycanHypothesisSerializer(glycan_file, db_file)
        glycan_builder.start()

        glycopeptide_builder = naive_glycopeptide.MultipleProcessFastaGlycopeptideHypothesisSerializer(
            fasta_file, db_file, glycan_builder.hypothesis_id,
            protease=['trypsin'],
            constant_modifications=constant_modifications,
            variable_modifications=[], max_missed_cleavages=2)
        glycopeptide_builder.start()

        peptides_without_uniprot = 80
        peptides_with_uniprot = 88
        peptides_with_uniprot_and_ragged_signal_peptide = 140

        peptides_count = glycopeptide_builder.query(serialize.Peptide).count()

        if peptides_count == peptides_with_uniprot or peptides_count == peptides_with_uniprot_and_ragged_signal_peptide:
            post_cleavage = glycopeptide_builder.query(serialize.Peptide).filter(
                serialize.Peptide.base_peptide_sequence == "DEASGIGPEEHFPEVPEIEPMGPVCPFR").first()
            self.assertIsNotNone(post_cleavage)
            self.assertEqual(
                len(glycopeptide_builder.query(serialize.Protein).first().annotations), 2
            )
        elif peptides_count == peptides_without_uniprot:
            warnings.warn("Failed to communicate with UniProt, skip this test")
        else:
            raise ValueError("Incorrect peptide count: %r" % (peptides_count, ))

        self.clear_file(db_file)
        self.clear_file(fasta_file)
        self.clear_file(glycan_file)

    def test_extract_forward_backward(self):
        fasta_file = fixtures.get_test_data("yeast_glycoproteins.fa")
        glycan_file = self.setup_tempfile(simple_n_glycans)
        forward_db = self.setup_tempfile("")
        reverse_db = self.setup_tempfile("")

        glycan_builder = TextFileGlycanHypothesisSerializer(glycan_file, forward_db)
        glycan_builder.start()

        builder = naive_glycopeptide.MultipleProcessFastaGlycopeptideHypothesisSerializer(
            fasta_file, forward_db, 1)
        cnt = builder.extract_proteins()
        assert cnt == 251

        glycan_builder = TextFileGlycanHypothesisSerializer(glycan_file, reverse_db)
        glycan_builder.start()

        rev_builder = naive_glycopeptide.ReversingMultipleProcessFastaGlycopeptideHypothesisSerializer(
            fasta_file, reverse_db, 1)
        cnt = rev_builder.extract_proteins()
        assert cnt == 251
        fwd_prots = builder.query(serialize.Protein).all()
        rev_prots = rev_builder.query(serialize.Protein).all()

        for fx, rx in zip(fwd_prots, rev_prots):
            assert fx.name == rx.name
            assert len(fx.glycosylation_sites) == len(rx.glycosylation_sites)


if __name__ == '__main__':
    unittest.main()



from collections import OrderedDict, namedtuple
import re
from typing import Dict, List, Tuple

import numpy as np

from scipy import sparse

from matplotlib import pyplot as plt

from ms_deisotope.feature_map.profile_transform import peak_indices
from glycresoft.composition_distribution_model.observation import GlycanCompositionSolutionRecord

from glycresoft.task import log_handle
from glycresoft.database.composition_network.graph import CompositionGraph

from .constants import DEFAULT_RHO
from .laplacian_smoothing import LaplacianSmoothingModel


class NetworkReduction(object):
    r"""A mapping-like ordered data structure that maps score thresholds to
    :class:`NetworkTrimmingSolution` instances produced at that score. Used
    for selecting the optimal smoothing factor :math:`\lambda` and during the
    grid search for finding the optimal score threshold.

    Attributes
    ----------
    store : :class:`~.OrderedDict`
        The storage mapping score to solution
    """

    store: OrderedDict
    _invalidated: bool

    def __init__(self, store=None):
        if store is None:
            store = OrderedDict()
        self.store = store
        self._invalidated = True
        self._resort()

    def getkey(self, key):
        """Get the solution whose score threshold matches ``key``

        Parameters
        ----------
        key : float
            The score threshold

        Returns
        -------
        :class:`NetworkTrimmingSolution`
        """
        return self.store[key]

    def getindex(self, ix):
        """Get the ``ix``th solution, regardless of threshold value

        Parameters
        ----------
        ix : int
            The index to return

        Returns
        -------
        :class:`NetworkTrimmingSolution`
        """
        return self.getkey(list(self.store.keys())[ix])

    def searchkey(self, value):
        """Search for the solution whose score threshold
        is closest to ``value``

        Parameters
        ----------
        value : float
            The threshold to search for

        Returns
        -------
        :class:`NetworkTrimmingSolution`
        """
        if self._invalidated:
            self._resort()
        array = list(self.store.keys())
        ix = self.binsearch(array, value)
        key = array[ix]
        return self.getkey(key)

    def put(self, key, value):
        self.store[key] = value
        self._invalidated = True

    def _resort(self):
        self.store = OrderedDict((key, self[key]) for key in sorted(self.store.keys()))
        self._invalidated = False

    def __getitem__(self, key):
        return self.getkey(key)

    def __setitem__(self, key, value):
        self.put(key, value)

    def __iter__(self):
        return iter(self.store.values())

    def __len__(self):
        return len(self.store)

    @staticmethod
    def binsearch(array, value):
        lo = 0
        hi = len(array)
        i = 0
        while (hi - lo) != 0:
            i = (hi + lo) // 2
            x = array[i]
            if abs(x - value) < 1e-3:
                return i
            elif (hi - lo) == 1:
                return i
            elif x < value:
                lo = i
            elif x > value:
                hi = i
        return i

    def plot(self, ax=None, **kwargs):
        if ax is None:
            fig, ax = plt.subplots(1)
        x = self.store.keys()
        y = [v.optimal_lambda for v in self.store.values()]
        ax.plot(x, y, **kwargs)
        xbound = ax.get_xlim()
        ybound = ax.get_ylim()
        ax.scatter(x, y)
        ax.set_xlim(*xbound)
        ax.set_ylim(*ybound)
        ax.set_xlabel(r"$S_o$ Threshold", fontsize=18)
        ax.set_ylabel(r"Optimal $\lambda$", fontsize=18)
        return ax

    def minimum_threshold_for_lambda(self, lmbda_target):
        best = None
        for value in reversed(self.store.values()):
            if best is None:
                best = value
                continue
            if abs(best.optimal_lambda - lmbda_target) >= abs(value.optimal_lambda - lmbda_target):
                if value.threshold < best.threshold:
                    best = value
        return best

    def press_weighted_mean_threshold(self):
        vals = list(self)
        return np.average(np.array(
            [v.threshold for v in vals]), weights=np.array(
            [v.press_residuals.min() for v in vals]))

    def keys(self):
        return self.store.keys()


class NetworkTrimmingSearchSolution(object):
    r"""Hold all the information for the grid search over the smoothing factor
    :math:`\lambda` at score threshold :attr:`threshold`.

    Attributes
    ----------
    lambda_values : :class:`np.ndarray`
        The :math:`\lambda` searched over
    minimum_residuals : float
        The smallest PRESS residuals
    model : :class:`~.LaplacianSmoothingModel`
        The smoothing model used for producing these estimates
    network : :class:`~.CompositionGraph`
        The network structure retained at :attr:`threshold`
    observed : list
        The graph nodes considered observed at this threshold
    optimal_lambda : float
        The best :math:`\lambda` value selected by minimizing the PRESS
    optimal_tau : :class:`np.ndarray`
        The value of :math:`\tau` for the best value of :math:`\lambda`
    press_residuals : :class:`np.ndarray`
        The PRESS at each value of :math:`\lambda`
    taus : list
        The :class:`np.ndarray` for each value of :math:`\lambda`
    threshold : float
        The score threshold used to produce this trimmed network
    updated : bool
        Whether or not this model has changed since the last threshold
    """

    threshold: float
    updated: bool
    observed: List[GlycanCompositionSolutionRecord]

    minimum_residuals: float
    press_residuals: np.ndarray

    optimal_lambda: float
    lambda_values: np.ndarray

    optimal_tau: np.ndarray
    taus: List[np.ndarray]

    model: LaplacianSmoothingModel
    network: CompositionGraph



    def __init__(self, threshold, lambda_values, press_residuals, network, observed=None,
                 updated=None, taus=None, model=None):
        self.threshold = threshold
        self.lambda_values = lambda_values
        self.press_residuals = press_residuals
        self.network = network
        self.taus = taus
        optimal_ix = np.argmin(self.press_residuals)
        self.optimal_lambda = self.lambda_values[optimal_ix]
        self.optimal_tau = self.taus[optimal_ix]
        self.minimum_residuals = self.press_residuals.min()
        self.observed = observed
        self.updated = updated
        self.model = model

    @property
    def n_kept(self):
        return len(self.network)

    @property
    def n_edges(self):
        return len(self.network.edges)

    @property
    def opt_lambda(self):
        return self.optimal_lambda

    def __repr__(self):
        min_press = min(self.press_residuals)
        opt_lambda = self.optimal_lambda
        return "NetworkTrimmingSearchSolution(%f, %d, %0.3f -> %0.3e)" % (
            self.threshold, self.n_kept, opt_lambda, min_press)

    def copy(self):
        dup = self.__class__(
            self.threshold,
            self.lambda_values.copy(),
            self.press_residuals.copy(),
            self.network.clone(),
            self.observed.copy(),
            list(self.updated),
            list(self.taus),
            self.model)
        return dup

    def plot(self, ax=None, **kwargs):
        if ax is None:
            fig, ax = plt.subplots(1)
        ax.plot(self.lambda_values, self.press_residuals, **kwargs)
        ax.set_xlabel(r"$\lambda$")
        ax.set_ylabel(r"Summed $PRESS$ Residual")
        return ax


GridSearchSolution = namedtuple("GridSearchSolution", (
    "tau_sequence", "tau_magnitude", "thresholds", "apexes",
    "target_thresholds"))



class GridPointTextCodec(object):
    version_code = 1

    def __init__(self, cls_type=None):
        if cls_type is None:
            cls_type = GridPointSolution
        self.cls_type = cls_type

    def load(self, fp):
        state = "BETWEEN"
        threshold = 0
        lmbda = 0
        tau = []
        belongingness_matrix = []
        variance_matrix_tags = {}
        neighborhood_names = []
        node_names = []
        version_code = self.version_code

        for line_number, line in enumerate(fp):
            line = line.strip("\n\r")
            if line.startswith(";version="):
                version_code = int(line.split("=", 1)[1])
            if line.startswith(";"):
                continue
            if line.startswith("threshold:"):
                threshold = float(line.split(":")[1])
                if state in ("TAU", "BELONG"):
                    state = "BETWEEN"
            elif line.startswith("lambda:"):
                lmbda = float(line.split(":")[1])
                if state in ("TAU", "BELONG"):
                    state = "BETWEEN"
            elif line.startswith("tau:"):
                state = "TAU"
            elif line.startswith("belongingness:"):
                state = "BELONG"
            elif line.startswith("variance:"):
                state = "VARIANCE"
            elif line.startswith("\t") or line.startswith("  "):
                if state == "TAU":
                    try:
                        _, name, value = re.split(r"\t|\s{2,}", line)
                    except ValueError as e:
                        print(line_number, line)
                        raise e
                    tau.append(float(value))
                    neighborhood_names.append(name)
                elif state == "BELONG":
                    try:
                        _, name, values = re.split(r"\t|\s{2,}", line)
                    except ValueError as e:
                        print(line_number, line)
                        raise e
                    belongingness_matrix.append([float(t) for t in values.split(",")])
                    node_names.append(name)
                elif state == 'VARIANCE':
                    try:
                        _, tag, values = re.split(r"\t|\s{2,}", line)
                        values = list(map(float, values.split(",")))
                        variance_matrix_tags[tag] = values
                    except ValueError as e:
                        print(line_number, line)
                        raise e
                else:
                    state = "BETWEEN"
        if variance_matrix_tags:
            vmat_shape = [int(p) for p in variance_matrix_tags['shape']]
            variance_matrix = np.zeros(vmat_shape)
            row_ix = np.array(variance_matrix_tags['rows'], dtype=int)
            col_ix = np.array(variance_matrix_tags['cols'], dtype=int)
            variance_matrix[row_ix, col_ix] = variance_matrix_tags['data']
        else:
            variance_matrix = None

        return self.cls_type(threshold, lmbda, np.array(tau, dtype=np.float64),
                   np.array(belongingness_matrix, dtype=np.float64),
                   neighborhood_names=neighborhood_names,
                   node_names=node_names, variance_matrix=variance_matrix)

    def dump(self, obj, fp):
        fp.write(";version=%s\n" % self.version_code)
        fp.write("threshold: %f\n" % (obj.threshold,))
        fp.write("lambda: %f\n" % (obj.lmbda,))
        fp.write("tau:\n")
        for i, t in enumerate(obj.tau):
            fp.write("\t%s\t%f\n" % (obj.neighborhood_names[i], t,))
        fp.write("belongingness:\n")
        for g, row in enumerate(obj.belongingness_matrix):
            fp.write("\t%s\t" % (obj.node_names[g]))
            for i, a_ij in enumerate(row):
                if i != 0:
                    fp.write(",")
                fp.write("%f" % (a_ij,))
            fp.write("\n")
        fp.write("variance:\n")
        var_coords = sparse.coo_matrix(obj.variance_matrix)
        fp.write('\tshape\t%s\n' % (','.join(map(str, var_coords.shape))))
        fp.write("\trow\t%s\n" % (','.join(map(str, var_coords.row))))
        fp.write("\tcol\t%s\n" % (','.join(map(str, var_coords.col))))
        fp.write("\tdata\t%s\n" % (','.join(map(str, var_coords.data))))
        return fp


class GridPointSolution(object):
    threshold: float
    lmbda: float

    tau: np.ndarray
    belongingness_matrix: np.ndarray
    variance_matrix: np.ndarray

    # Label arrays
    neighborhood_names: np.ndarray
    node_names: np.ndarray

    def __init__(self, threshold, lmbda, tau, belongingness_matrix, neighborhood_names, node_names,
                 variance_matrix=None):
        if variance_matrix is None:
            variance_matrix = np.identity(len(node_names), dtype=np.float64)
        self.threshold = threshold
        self.lmbda = lmbda
        self.tau = np.array(tau, dtype=np.float64)
        self.belongingness_matrix = np.array(belongingness_matrix, dtype=np.float64)
        self.neighborhood_names = np.array(neighborhood_names)
        self.node_names = np.array(node_names)
        self.variance_matrix = np.array(variance_matrix, dtype=np.float64)

    def __eq__(self, other):
        ac = np.allclose
        match = ac(self.threshold, other.threshold) and ac(self.lmbda, other.lmbda) and ac(
            self.tau, other.tau) and ac(self.belongingness_matrix, other.belongingness_matrix) and ac(
            self.variance_matrix, other.variance_matrix)
        if not match:
            return match
        match = np.all(self.node_names == other.node_names) and np.all(
            self.neighborhood_names == other.neighborhood_names)
        return match

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "GridPointSolution(threshold=%0.3f, lmbda=%0.3f, tau=%r)" % (
            self.threshold, self.lmbda, self.tau)

    def clone(self):
        return self.__class__(
            self.threshold,
            self.lmbda,
            self.tau.copy(),
            self.belongingness_matrix.copy(),
            self.neighborhood_names.copy(),
            self.node_names.copy(),
            self.variance_matrix.copy())

    def reindex(self, model: LaplacianSmoothingModel):
        node_indices, node_names = self._build_node_index_map(model)
        tau_indices = self._build_neighborhood_index_map(model)

        self.belongingness_matrix = self.belongingness_matrix[node_indices, :][:, tau_indices]
        self.tau = self.tau[tau_indices]
        self.variance_matrix = self.variance_matrix[node_indices, node_indices]

        self.neighborhood_names = model.neighborhood_names
        self.node_names = node_names

    def _build_node_index_map(self, model: LaplacianSmoothingModel) -> Tuple[List[int], List[str]]:
        name_to_new_index: Dict[str, int] = dict()
        name_to_old_index: Dict[str, int] = dict()
        for i, node_name in enumerate(self.node_names):
            name_to_old_index[node_name] = i
            name_to_new_index[node_name] = model.network[node_name].index
        assert len(name_to_new_index) == len(self.node_names)
        ordering = [None for i in range(len(self.node_names))]
        new_name_order = [None for i in range(len(self.node_names))]
        for name, new_index in name_to_new_index.items():
            old_index = name_to_old_index[name]
            ordering[new_index] = old_index
            new_name_order[new_index] = name
        for x in ordering:
            assert x is not None
        return ordering, new_name_order

    def _build_neighborhood_index_map(self, model: LaplacianSmoothingModel) -> List[int]:
        tau_indices = [model.neighborhood_names.index(name) for name in self.neighborhood_names]
        return tau_indices

    def dump(self, fp, codec=None):
        if codec is None:
            codec = GridPointTextCodec()
        codec.dump(self, fp)
        return fp

    @classmethod
    def load(cls, fp, codec=None):
        if codec is None:
            codec = GridPointTextCodec(cls)
        inst = codec.load(fp)
        return inst


class NetworkSmoothingModelSolutionBase(object):
    def __init__(self, model):
        self.model = model

    def _get_default_solution(self, *args, **kwargs):
        raise NotImplementedError()

    def estimate_phi_observed(self, solution=None, remove_threshold=True, rho=DEFAULT_RHO):
        if solution is None:
            solution = self._get_default_solution(rho=rho)
        if remove_threshold:
            self.model.reset()
        return self.model.optimize_observed_scores(
            solution.lmbda, solution.belongingness_matrix[self.model.obs_ix, :].dot(solution.tau))

    def estimate_phi_missing(self, solution=None, remove_threshold=True, observed_scores=None):
        if solution is None:
            solution = self._get_default_solution()
        if remove_threshold:
            self.model.reset()
        if observed_scores is None:
            observed_scores = self.estimate_phi_observed(
                solution=solution, remove_threshold=False)
        t0 = self.model.A0.dot(solution.tau)
        tm = self.model.Am.dot(solution.tau)
        return self.model.compute_missing_scores(observed_scores, t0, tm)

    def annotate_network(self, solution=None, remove_threshold=True, include_missing=True):
        if solution is None:
            solution = self._get_default_solution()
        if remove_threshold:
            self.model.reset()
        observed_scores = self.estimate_phi_observed(solution, remove_threshold=False)

        if include_missing:
            missing_scores = self.estimate_phi_missing(
                solution, remove_threshold=False,
                observed_scores=observed_scores)

        network = self.model.network.clone()
        network.neighborhoods = self.model.neighborhood_walker.neighborhoods.clone()
        for i, ix in enumerate(self.model.obs_ix):
            network[ix].score = observed_scores[i]

        if include_missing:
            for i, ix in enumerate(self.model.miss_ix):
                network[ix].score = missing_scores[i]
                network[ix].marked = True

        return network


class ThresholdSelectionGridSearch(NetworkSmoothingModelSolutionBase):
    def __init__(self, model, network_reduction=None, apex_threshold=0.95, threshold_bias=4.0):
        super(ThresholdSelectionGridSearch, self).__init__(model)
        self.network_reduction = network_reduction
        self.apex_threshold = apex_threshold
        self.threshold_bias = float(threshold_bias)
        if self.threshold_bias < 1:
            raise ValueError("Threshold Bias must be 1 or greater")

    def _get_default_solution(self, *args, **kwargs):
        return self.average_solution(*args, **kwargs)

    def has_reduction(self):
        return self.network_reduction is not None and bool(self.network_reduction)

    def explore_grid(self):
        if self.network_reduction is None:
            self.network_reduction = self.model.find_threshold_and_lambda(
                rho=DEFAULT_RHO, threshold_step=0.1, fit_tau=True)
        log_handle.log("... Exploring Grid Landscape")
        stack = []
        tau_magnitude = []
        thresholds = []

        for level in self.network_reduction:
            thresholds.append(level.threshold)

        # Pull the distribution slightly to the right
        bias_shift = 1 - (1 / self.threshold_bias)
        # Reduces the influence of the threshold
        bias_scale = self.threshold_bias

        for level in self.network_reduction:
            stack.append(np.array(level.taus).mean(axis=0))
            tau_magnitude.append(
                np.abs(level.optimal_tau).sum() * (
                    (level.threshold / bias_scale) + bias_shift)
            )
        tau_magnitude = np.array(tau_magnitude)
        if len(tau_magnitude) == 0:
            # No solutions, so these will be empty
            return GridSearchSolution(stack, tau_magnitude, thresholds, np.array([]), thresholds)
        elif len(tau_magnitude) <= 2:
            apex = np.argmax(tau_magnitude)
        elif len(tau_magnitude) > 2:
            apex = peak_indices(tau_magnitude)
            if len(apex) == 0:
                apex = np.array([np.argmax(tau_magnitude)])

        thresholds = np.array(thresholds)
        apex_threshold = tau_magnitude[apex].max() * self.apex_threshold
        if apex_threshold != 0:
            apex = apex[(tau_magnitude[apex] > apex_threshold)]
        else:
            # The tau threshold may be 0, in which case any point will do, but this
            # solution carries no generalization.
            apex = apex[(tau_magnitude[apex] >= apex_threshold)]
        target_thresholds = [t for t in thresholds[apex]]
        solution = GridSearchSolution(stack, tau_magnitude, thresholds, apex, target_thresholds)
        log_handle.log("... %d Candidate Solutions" % (len(target_thresholds),))
        return solution

    def _get_solution_states(self):
        solution = self.explore_grid()
        states = []
        for i, t in enumerate(solution.target_thresholds):
            states.append(self.network_reduction.searchkey(t))
        return states

    def _get_estimate_for_state(self, state, rho=DEFAULT_RHO, lmbda=None):
        # Get optimal value of lambda based on
        # the PRESS for this group
        if lmbda is None:
            i = np.argmin(state.press_residuals)
            lmbda = state.lambda_values[i]

        # Removes rows from A0
        self.model.set_threshold(state.threshold)

        tau = self.model.estimate_tau_from_S0(rho, lmbda)
        A = self.model.normalized_belongingness_matrix.copy()

        # Restore removed rows from A0
        self.model.remove_belongingness_patch()

        variance_matrix = np.identity(len(self.model.network)) * 0
        variance_matrix[self.model.obs_ix, self.model.obs_ix] = np.diag(self.model.variance_matrix)

        return GridPointSolution(state.threshold, lmbda, tau, A,
                                 self.model.neighborhood_names,
                                 self.model.node_names,
                                 variance_matrix=variance_matrix)

    def get_solutions(self, rho=DEFAULT_RHO, lmbda=None):
        states = self._get_solution_states()
        solutions = [self._get_estimate_for_state(
            state, rho=rho, lmbda=lmbda) for state in states]
        self.model.reset()
        return solutions

    def average_solution(self, rho=DEFAULT_RHO, lmbda=None):
        solutions = self.get_solutions(rho=rho, lmbda=lmbda)
        if len(solutions) == 0:
            return None
        tau_acc = np.zeros_like(solutions[0].tau)
        lmbda_acc = 0
        thresh_acc = 0
        # variance_matrix = np.zeros_like()
        A = np.zeros_like(solutions[0].belongingness_matrix)
        for sol in solutions:
            thresh_acc += sol.threshold
            tau_acc += sol.tau
            lmbda_acc += sol.lmbda
            A += sol.belongingness_matrix
        n = len(solutions)
        thresh_acc /= n
        tau_acc /= n
        lmbda_acc /= n
        A /= n
        # A = ProportionMatrixNormalization.normalize(A, self.model._belongingness_normalization)
        variance_matrix = np.identity(len(self.model.network)) * 0
        variance_matrix[self.model.obs_ix, self.model.obs_ix] = np.diag(self.model.variance_matrix)
        average_solution = GridPointSolution(
            thresh_acc, lmbda_acc, tau_acc, A,
            self.model.neighborhood_names, self.model.node_names,
            variance_matrix=variance_matrix)
        return average_solution

    def plot(self, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1)
        solution = self.explore_grid()
        ax.plot(solution.thresholds, solution.tau_magnitude)
        ax.scatter(
            solution.thresholds[solution.apexes],
            solution.tau_magnitude[solution.apexes])
        ax.set_xlabel("Threshold", fontsize=18)
        ax.set_ylabel("Criterion", fontsize=18)
        ax.set_title("Locate Ideal Threshold\nBy Maximizing ${\\bar \\tau_j}$", fontsize=28)
        ax.set_xticks([x_ for i, x_ in enumerate(solution.thresholds) if i % 5 == 0])
        ax.set_xticklabels(["%0.2f" % x_ for i, x_ in enumerate(solution.thresholds) if i % 5 == 0], rotation=90)
        return ax

    def plot_thresholds(self, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1)
        solution = self.explore_grid()
        ax = self.network_reduction.plot(ax)
        ax.vlines(solution.thresholds[solution.apexes], 0, 1, color='red')
        ax.set_title("Selected Estimation Points for ${\\bar \\tau}$", fontsize=28)
        return ax

    def fit(self, rho=DEFAULT_RHO, lmbda=None):
        solution = self.average_solution(rho=rho, lmbda=lmbda)
        if solution is None:
            raise ValueError("Could not fit model. No acceptable solution found.")
        return NetworkSmoothingModelFit(self.model, solution)


class NetworkSmoothingModelFit(NetworkSmoothingModelSolutionBase):
    def __init__(self, model, solution):
        super(NetworkSmoothingModelFit, self).__init__(model)
        self.solution = solution

    def _get_default_solution(self, *args, **kwargs):
        return self.solution

    def __repr__(self):
        return "{self.__class__.__name__}({self.model}, {self.solution})".format(self=self)



import os
import time
import traceback
from typing import Any, List, Mapping, Tuple, Type


try:
    import cPickle as pickle
except ImportError:
    import pickle

from collections import deque
from threading import Thread

import multiprocessing
from multiprocessing import Process, Event, Manager, JoinableQueue
from multiprocessing.managers import RemoteError

try:
    from Queue import Empty as QueueEmptyException
except ImportError:
    from queue import Empty as QueueEmptyException

from glypy.utils import uid

from ms_deisotope.data_source import ProcessedScan

from glycresoft.task import TaskBase
from glycresoft.chromatogram_tree import Unmodified, MassShift
from glycresoft.structure import LRUMapping

from .evaluation import SolutionHandler, LocalSpectrumEvaluator, SpectrumEvaluatorBase
from .task import TaskQueueFeeder, DBHit, HitID, GroupID
from .utils import SentinelToken, ProcessDispatcherState


debug_mode = bool(os.environ.get("GLYCRESOFTDEBUG"))


class IdentificationProcessDispatcher(TaskBase):
    """Orchestrates distributing the spectrum match evaluation
    task across several processes.

    The distribution pushes individual structures ("targets") and
    the scan ids they mapped to in the MSn dimension to each worker.

    All scans in the batch being worked on are made available over
    an IPC synchronized dictionary.

    Attributes
    ----------
    done_event : multiprocessing.Event
        An Event indicating that all work items
        have been placed on `input_queue`
    evaluation_args : dict
        A dictionary containing arguments to be
        passed through to `evaluate` on the worker
        processes.
    init_args : dict
        A dictionary containing extra arguments
        to use when initializing worker process
        instances.
    input_queue : multiprocessing.Queue
        The queue from which worker processes
        will read their targets and scan id mappings
    ipc_manager : multiprocessing.SyncManager
        Provider of IPC dictionary synchronization
    log_controller : MessageSpooler
        Logging facility to funnel messages from workers
        through into the main process's log stream
    n_processes : int
        The number of worker processes to spawn
    output_queue : multiprocessing.Queue
        The queue which worker processes will
        put ther results on, read in the main
        process.
    scan_load_map : multiprocessing.SyncManager.dict
        An inter-process synchronized dictionary which
        maps scan ids to scans. Used by worker processes
        to request individual scans by name when they are
        not found locally.
    scan_solution_map : defaultdict(list)
        A mapping from scan id to all candidate solutions.
    scorer_type : SpectrumMatcherBase
        The type used by workers to evaluate spectrum matches
    worker_type : SpectrumIdentificationWorkerBase
        The type instantiated to construct worker processes
    workers : list
        Container for created workers.
    """

    solution_handler: SolutionHandler
    state: ProcessDispatcherState
    worker_type: Type['SpectrumIdentificationWorkerBase']

    input_queue: JoinableQueue
    output_queue: JoinableQueue

    producer_thread_done_event: Event


    post_search_trailing_timeout = 1.5e2
    child_failure_timeout = 2.5e2

    def __init__(self, worker_type, scorer_type, evaluation_args=None, init_args=None,
                 mass_shift_map=None, n_processes=3, ipc_manager=None, solution_handler_type=None):
        if solution_handler_type is None:
            solution_handler_type = SolutionHandler
        if ipc_manager is None:
            self.log("Creating IPC Manager. Prefer to pass a reusable IPC Manager instead.")
            ipc_manager = Manager()
        if evaluation_args is None:
            evaluation_args = dict()
        if mass_shift_map is None:
            mass_shift_map = {
                Unmodified.name: Unmodified
            }

        self.state = ProcessDispatcherState.start
        self.ipc_manager = ipc_manager
        self.worker_type = worker_type
        self.scorer_type = scorer_type
        self.n_processes = n_processes

        self.producer_thread_done_event = self.ipc_manager.Event()
        self.consumer_done_event = self.ipc_manager.Event()

        self.input_queue = self._make_input_queue()
        self.output_queue = self._make_output_queue()

        self.feeder = TaskQueueFeeder(self.input_queue, self.producer_thread_done_event)

        self.solution_handler = solution_handler_type({}, {}, mass_shift_map)

        self.evaluation_args = evaluation_args
        self.init_args = init_args
        self.workers = []
        self.log_controller = self.ipc_logger()
        self.local_scan_map = dict()
        self.scan_load_map = self.ipc_manager.dict()
        self.local_mass_shift_map = mass_shift_map
        self.mass_shift_load_map = self.ipc_manager.dict(mass_shift_map)
        self.structure_map = dict()
        self._token_to_worker = {}
        self._has_received_token = set()
        self._has_remote_error = False
        self._result_buffer = deque()

    @property
    def scan_solution_map(self):
        return self.solution_handler.scan_solution_map

    def _make_input_queue(self):
        try:
            return JoinableQueue(int(1e5))
        except (OSError, ValueError):
            return JoinableQueue()

    def _make_output_queue(self):
        try:
            return JoinableQueue(int(1e7))
        except (OSError, ValueError):
            return JoinableQueue()

    def clear_pool(self):
        """Tear down spawned worker processes and clear
        the shared memory server
        """
        self.scan_load_map.clear()
        self.local_scan_map.clear()
        if self.state in (ProcessDispatcherState.running, ProcessDispatcherState.running_local_workers_dead):
            self.state = ProcessDispatcherState.terminating
        elif self.state == ProcessDispatcherState.running_local_workers_live:
            self.state = ProcessDispatcherState.terminating_workers_live
        else:
            self.state = ProcessDispatcherState.terminating
        for _i, worker in enumerate(self.workers):
            exitcode = worker.exitcode
            if exitcode != 0 and exitcode is not None:
                self.log("... Worker Process %r had exitcode %r" % (worker, exitcode))
            try:
                worker.join(1)
            except AttributeError:
                pass
            if worker.is_alive() and worker.token not in self._has_received_token:
                self.debug("... Worker Process %r is still alive and incomplete" % (worker, ))
                worker.terminate()

    def create_pool(self, scan_map):
        """Spawn a pool of workers and a supporting process
        for sharing scans from ``scan_map`` by id with the workers
        so they can load scans on demand.

        Parameters
        ----------
        scan_map : dict
            Map scan id to :class:`.ProcessedScan` object
        """
        self.state = ProcessDispatcherState.spawning
        self.scan_load_map.clear()
        # Do not copy the complete scan object into the IPC manager dict, as this will
        # pickle the scan source, and then unpickle it several times on the recieving end.
        self.scan_load_map.update({k: pickle.dumps(v.copy(deep=False).unbind(), -1) for k, v in scan_map.items()})
        self.local_scan_map.clear()
        self.local_scan_map.update(scan_map)

        self.input_queue = self._make_input_queue()
        self.output_queue = self._make_output_queue()
        self.feeder = TaskQueueFeeder(self.input_queue, self.producer_thread_done_event)

        for _i in range(self.n_processes):
            worker = self.worker_type(
                input_queue=self.input_queue,
                output_queue=self.output_queue,
                producer_done_event=self.producer_thread_done_event,
                consumer_done_event=self.consumer_done_event,
                scorer_type=self.scorer_type,
                evaluation_args=self.evaluation_args,
                spectrum_map=self.scan_load_map,
                mass_shift_map=self.mass_shift_load_map,
                log_handler=self.log_controller.sender(),
                solution_packer=self.solution_handler.packer,
                **self.init_args)
            worker._work_complete = self.ipc_manager.Event()
            worker.start()
            self._token_to_worker[worker.token] = worker
            self.workers.append(worker)

    def all_workers_finished(self):
        """Check if all worker processes have finished.
        """
        worker_still_busy = False
        i = 0
        j = 0
        for worker in self.workers:
            i += 1
            try:
                is_done = worker.all_work_done()
                j += is_done
                if not is_done:
                    worker_still_busy = True
                    return worker_still_busy
            except (RemoteError, KeyError):
                worker_still_busy = True
                self._has_remote_error = True
        self.log("... All Workers Done: %r (%d/%d), Error? %r" % (not worker_still_busy, j, i, self._has_remote_error))
        return not worker_still_busy

    def build_work_order(self, hit_id: HitID, hit_map: Mapping[HitID, DBHit],
                         scan_hit_type_map: Mapping[Tuple[str, HitID], MassShift],
                         hit_to_scan: Mapping[HitID, List[str]]):
        """Packs several task-defining data structures into a simple to unpack payload for
        sending over IPC to worker processes.

        Parameters
        ----------
        hit_id : int
            The id number of a hit structure
        hit_map : dict
            Maps hit_id to hit structure
        hit_to_scan : dict
            Maps hit id to list of scan ids
        scan_hit_type_map : dict
            Maps (hit id, scan id) to the type of mass shift
            applied for this match

        Returns
        -------
        tuple
            Packaged message payload
        """
        return self.feeder.build_work_order(hit_id, hit_map, scan_hit_type_map, hit_to_scan)

    def spawn_queue_feeder(self, hit_map, hit_to_scan, scan_hit_type_map, hit_to_group=None):
        """Create a thread to run :meth:`feeder` with the provided arguments
        so that work can be sent in tandem with waiting for results

        Parameters
        ----------
        hit_map : dict
            Maps hit id to structure
        hit_to_scan : dict
            Maps hit id to list of scan ids
        scan_hit_type_map : dict
            Maps (hit id, scan id) to the type of mass shift
            applied for this match
        hit_to_group: dict, optional
            Maps group id to the set of hit ids which are
        Returns
        -------
        Thread
        """
        feeder_thread = Thread(
            target=self.feeder, args=(hit_map, hit_to_scan, scan_hit_type_map, hit_to_group))
        feeder_thread.daemon = True
        feeder_thread.start()
        return feeder_thread

    def store_result(self, target, score_map):
        """Save the spectrum match scores for ``target`` against the
        set of matched scans

        Parameters
        ----------
        target : object
            The structure that was matched
        score_map : dict
            Maps (scan id, mass shift name) to score
        scan_map : dict
            Maps scan id to :class:`.ProcessedScan`
        """
        self.solution_handler(target, score_map)

    def _reconstruct_missing_work_items(self, seen, hit_map, hit_to_scan, scan_hit_type_map):
        """Handle task items that are pending after it is believed that the workers
        have crashed or the network communication has failed.

        Parameters
        ----------
        seen : dict
            Map of hit ids that have already been handled
        hit_map : dict
            Maps hit_id to hit structure
        hit_to_scan : dict
            Maps hit id to list of scan ids
        scan_hit_type_map : dict
            Maps (hit id, scan id) to the type of mass shift
            applied for this match
        """
        missing = set(hit_to_scan) - set(seen)
        i = 0
        n = len(missing)
        evaluator = LocalSpectrumEvaluator(
            self.workers[0], self.local_scan_map, self.local_mass_shift_map,
            self.solution_handler.packer)
        for missing_id in missing:
            target, scan_spec = self.build_work_order(missing_id, hit_map, scan_hit_type_map, hit_to_scan)
            try:
                # Change this to be the target's ID and look it up in hit_map
                target_id, score_map = evaluator.handle_item(target, scan_spec)
                target = hit_map.get(target_id, target_id)
                seen[target.id] = (-1, 0)
                self.store_result(target, score_map)
                i += 1
                if i % 1000 == 0:
                    self.log("...... Processed %d local matches (%0.2f%%)" % (i, i * 100. / n))
            except Exception as err:
                self.error("...... An error occurred while processing %r: %r" % (target, err), exception=err)
                raise
        return i

    def get_result(self, seen, hit_map):
        if self._result_buffer:
            return self._result_buffer.popleft()

        payload = self.output_queue.get(True, 1)
        self.output_queue.task_done()

            # Change this to be the target's ID and look it up in hit_map
        if isinstance(payload, SentinelToken):
            self.debug("...... Received sentinel from %s" %
                        (self._token_to_worker[payload.token].name))
            self._has_received_token.add(payload.token)
        else:
            self._result_buffer.extend(payload)
            if self._result_buffer:
                return self._result_buffer.popleft()

    def process(self, scan_map: Mapping[str, ProcessedScan],
                hit_map: Mapping[HitID, DBHit],
                hit_to_scan: Mapping[HitID, List[str]],
                scan_hit_type_map: Mapping[Tuple[str, HitID], MassShift],
                hit_group_map: Mapping[GroupID, List[HitID]]=None) -> Mapping[str, List]:
        """Evaluate all spectrum matches, spreading work among the worker
        process pool.

        Parameters
        ----------
        scan_map : :class:`dict`
            Mapping from :attr:`~.ScanBase.id` to :class:`~.ScanBase` instance
        hit_map : :class:`dict`
            Mapping from `structure.id` to the structure type to matche against
        hit_to_scan : :class:`dict`
            Mapping from `structure.id` to a list of :attr:`~.ScanBase.id` which
            it has a precursor mass match.
        scan_hit_type_map : :class:`dict`
            Mapping from (:attr:`~.ScanBase.id`, `structure.id`) to :class:`MassShiftBase`
        hit_group_map : :class:`dict`, optional
            An additional :class:`dict` that maps abstract group ids to sets of hit ids
            in `hit_map` which all share information that can be computed once and shared
            between all members of the group.
        """
        self.structure_map = hit_map
        self.solution_handler.scan_map = scan_map
        self.create_pool(scan_map)

        # Won't feed hit groups until after more work is done here.
        feeder_thread = self.spawn_queue_feeder(
            hit_map, hit_to_scan, scan_hit_type_map, hit_group_map)
        has_work = True
        i = 0
        scan_count = len(scan_map)
        n = len(hit_to_scan)
        if n != len(hit_map):
            self.log("There is a mismatch between hit_map (%d) and hit_to_scan (%d)" % (
                len(hit_map), n))
        n_spectrum_matches = sum(map(len, hit_to_scan.values()))
        # Track the iteration number a particular structure (id) has been received
        # on. This may be used to detect if a structure has been received multiple
        # times, and to determine when all expected structures have been received.
        seen = dict()
        # Keep a running tally of the number of iterations when there are pending
        # structure matches to process, but all workers claim to be done.
        strikes = 0
        self.state = ProcessDispatcherState.running
        self.log("... Searching Matches (%d)" % (n_spectrum_matches,))
        last_log_time = start_time = time.time()
        should_log = False
        log_cycle = 5000
        while has_work:
            try:
                payload = self.get_result(seen, hit_map)
                if payload is None:
                    continue
                else:
                    (target_id, score_map, token) = payload
                target = hit_map.get(target_id, target_id)
                if target.id in seen:
                    self.log(
                        "...... Duplicate Results For %s. First seen at %r, now again at %r" % (
                            target, seen[target.id], (i, token)))
                else:
                    seen[target.id] = (i, token)
                if (i > n) and ((i - n) % 10 == 0):
                    self.log(
                        "...... Warning: %d additional output received. %s and %d matches." % (
                            i - n, target, len(score_map)))

                i += 1
                strikes = 0
                if i % log_cycle == 0 or should_log:
                    last_log_time = time.time()
                    should_log = False
                    self.log(
                        "...... Processed %d structures (%0.2f%%)" % (i, i * 100. / n))
                self.store_result(target, score_map)
            except QueueEmptyException:
                if len(seen) == n:
                    has_work = False
                # do worker life cycle management here
                elif self.all_workers_finished():
                    if len(seen) == n:
                        has_work = False
                    else:
                        strikes += 1
                        if strikes % 50 == 0:
                            self.log(
                                "...... %d cycles without output (%d/%d, %0.2f%% Done)" % (
                                    strikes, len(seen), n, len(seen) * 100. / n))
                        if strikes > self.post_search_trailing_timeout:
                            self.state = ProcessDispatcherState.running_local_workers_dead
                            self.log(
                                "...... Too much time has elapsed with"
                                " missing items. Evaluating serially.")
                            i += self._reconstruct_missing_work_items(
                                seen, hit_map, hit_to_scan, scan_hit_type_map)
                            has_work = False
                            self.debug("...... Processes")
                            for worker in self.workers:
                                self.debug("......... %r" % (worker,))
                            self.debug("...... IPC Manager: %r" % (self.ipc_manager,))
                else:
                    strikes += 1
                    if strikes % 10 == 0:
                        check_time = time.time()
                        if check_time - last_log_time > 10:
                            should_log = True
                    if strikes % 50 == 0:
                        self.log(
                            "...... %d cycles without output (%d/%d, %0.2f%% Done, %d children still alive)" % (
                                strikes, len(seen), n, len(seen) * 100. / n,
                                len(multiprocessing.active_children()) - 1))
                        try:
                            input_queue_size = self.input_queue.qsize()
                        except Exception:
                            input_queue_size = -1
                        is_feeder_done = self.producer_thread_done_event.is_set()
                        self.log("...... Input Queue Status: %r. Is Feeder Done? %r" % (
                            input_queue_size, is_feeder_done))
                    if strikes > (self.child_failure_timeout * (1 + (scan_count / 500.0) * (
                            not self._has_remote_error))):
                        self.state = ProcessDispatcherState.running_local_workers_live
                        self.log(
                            ("...... Too much time has elapsed with"
                             " missing items (%d children still alive). Evaluating serially.") % (
                                 len(multiprocessing.active_children()) - 1,))
                        i += self._reconstruct_missing_work_items(
                            seen, hit_map, hit_to_scan, scan_hit_type_map)
                        has_work = False
                        self.debug("...... Processes")
                        for worker in self.workers:
                            self.debug("......... %r" % (worker,))
                        self.debug("...... IPC Manager: %r" % (self.ipc_manager,))
                continue

        consumer_end = time.time()
        self.debug("... Consumer Done (%0.3g sec.)" % (consumer_end - start_time))
        self.consumer_done_event.set()
        time.sleep(1) # Not a good solution but need to give workers a chance to sync
        i_spectrum_matches = sum(map(len, self.scan_solution_map.values()))
        self.log("... Finished Processing Matches (%d)" % (i_spectrum_matches,))
        self.clear_pool()
        self.debug("... Shutting Down Message Queue")
        self.log_controller.stop()
        self.debug("... Joining Feeder Thread (Done: %r)" % (self.producer_thread_done_event.is_set(), ))
        feeder_thread.join()
        dispatcher_end = time.time()
        self.log("... Dispatcher Finished (%0.3g sec.)" % (dispatcher_end - start_time))
        return self.scan_solution_map


class SpectrumIdentificationWorkerBase(Process, SpectrumEvaluatorBase):
    verbose = False

    def __init__(self, input_queue, output_queue, producer_done_event, consumer_done_event,
                 scorer_type, evaluation_args, spectrum_map, mass_shift_map, log_handler,
                 solution_packer):
        Process.__init__(self)
        if evaluation_args is None:
            evaluation_args = dict()
        self.daemon = True
        self.input_queue = input_queue
        self.output_queue = output_queue
        self.producer_done_event = producer_done_event
        self.consumer_done_event = consumer_done_event
        self.scorer_type = scorer_type
        self.evaluation_args = evaluation_args

        self.solution_packer = solution_packer

        self.spectrum_map = spectrum_map
        self.mass_shift_map = mass_shift_map

        self.local_scan_map = LRUMapping(1000)
        self.local_mass_shift_map = dict({
            Unmodified.name: Unmodified
        })
        self.solution_map = dict()
        self._work_complete = Event()
        self._work_complete.clear()
        self.log_handler = log_handler
        self.token = uid()
        self.items_handled = 0
        self.result_buffer = []
        self.buffer_size = 1000
        self.last_sent_result = time.time()

    def log(self, message):
        """Send a normal logging message via :attr:`log_handler`

        Parameters
        ----------
        message : str
            The message to log
        """
        if self.log_handler is not None:
            self.log_handler(message)

    def debug(self, message):
        """Send a debugging message via :attr:`log_handler`

        Parameters
        ----------
        message : str
            The message to log
        """
        if self.verbose:
            self.log_handler("DEBUG::%s" % message)

    def fetch_scan(self, key):
        try:
            return self.local_scan_map[key]
        except KeyError:
            serialized_scan = self.spectrum_map[key]
            scan = pickle.loads(serialized_scan)
            self.local_scan_map[key] = scan
            return scan

    def fetch_mass_shift(self, key):
        try:
            return self.local_mass_shift_map[key]
        except KeyError:
            mass_shift = self.mass_shift_map[key]
            self.local_mass_shift_map[key] = mass_shift
            return mass_shift

    def all_work_done(self):
        """A helper method to encapsulate :attr:`_work_complete`'s ``is_set``
        method.

        Returns
        -------
        bool
        """
        return self._work_complete.is_set()

    def _append_to_result_buffer(self, payload):
        self.result_buffer.append(payload)
        if len(self.result_buffer) > self.buffer_size:
            self._flush_result_buffer()

    def _flush_result_buffer(self):
        if self.result_buffer:
            self.output_queue.put(self.result_buffer)
            self.result_buffer = []
            self.last_sent_result = time.time()

    def pack_output(self, target):
        """Transmit the completed identifications for a given target
        structure.

        Rather than returning these values directly, this implementation uses
        the output queue to send the target and its scores along an IPC queue.

        Parameters
        ----------
        target : object
            The structure being identified.
        """
        if self.solution_map:
            target_id = getattr(target, 'id', target)
            self._append_to_result_buffer((target_id, self.solution_map, self.token))
        self.solution_map = dict()

    def evaluate(self, scan, structure, evaluation_context=None, *args, **kwargs):
        raise NotImplementedError()

    def cleanup(self):
        """Send signals indicating the worker process is finished and do any
        final shared resource cleanup needed on the worker's side.

        This will set the :attr:`_work_complete` event and join :attr:`output_queue`
        """
        self.debug("... Process %s Setting Work Complete Flag. Processed %d structures" % (
            self.name, self.items_handled))
        try:
            self._work_complete.set()
        except (RemoteError, KeyError):
            self.log("An error occurred while cleaning up worker %r" % (self, ))
        self._flush_result_buffer()
        self.output_queue.put(SentinelToken(self.token))
        self.consumer_done_event.wait()
        # joining the queue may not be necessary if we depend upon consumer_event_done
        self.debug("... Process %s Queue Joining" % (self.name,))
        self.output_queue.join()
        self.debug("... Process %s Finished" % (self.name,))

    def before_task(self):
        '''A method to be overriden by subclasses that want to do something before
        starting the task loop.
        '''
        pass

    def task(self):
        """The worker process's main loop where it will poll for new work items,
        process incoming work items and send them back to the master process.
        """
        has_work = True
        self.items_handled = 0
        strikes = 0
        while has_work:
            try:
                payload = self.input_queue.get(True, 5)
                self.input_queue.task_done()
                strikes = 0
            except QueueEmptyException:
                if self.producer_done_event.is_set():
                    has_work = False
                    break
                else:
                    strikes += 1
                    self._flush_result_buffer()
                    if strikes % 1000 == 0:
                        self.log("... %d iterations without work for %r" % (strikes, self))
                    continue
            if self.items_handled % 100 == 0:
                if time.time() - self.last_sent_result > 60:
                    self._flush_result_buffer()
            # Handling a group of work items
            if isinstance(payload, dict):
                work_order = payload
                self.items_handled += 1
                try:
                    self.handle_group(work_order)
                except Exception:
                    message = "An error occurred while processing %r on %r:\n%s" % (
                        work_order, self, traceback.format_exc())
                    self.log(message)
                    break
            else: # Handling a single work item
                structure, scan_ids = payload
                self.items_handled += 1
                try:
                    self.handle_item(structure, scan_ids)
                except Exception:
                    message = "An error occurred while processing %r on %r:\n%s" % (
                        structure, self, traceback.format_exc())
                    self.log(message)
                    break
        self.cleanup()

    def run(self):
        new_name = getattr(self, 'process_name', None)
        if new_name is not None:
            TaskBase().try_set_process_name(new_name)
        try:
            self.before_task()
        except Exception:
            self.log("An exception occurred during before_task for %r.\n%s" % (
                self, traceback.format_exc()))
        try:
            self.task()
        except Exception:
            self.log("An exception occurred while executing %r.\n%s" % (
                self, traceback.format_exc()))
            self.cleanup()



import re
import os
from functools import partial

from typing import Optional, TYPE_CHECKING

from sqlalchemy import (
    Column, Integer, String, ForeignKey, PickleType,
    func)

from sqlalchemy.orm import relationship, backref, object_session
from sqlalchemy.ext.declarative import declared_attr

import dill

from glycresoft.serialize.base import (
    Base, HasUniqueName)

from glycresoft.serialize.spectrum import SampleRun

from glycresoft.serialize.hypothesis.generic import HasFiles
from glycresoft.serialize.param import HasParameters

from ms_deisotope.data_source import ProcessedRandomAccessScanSource
from ms_deisotope.output import ProcessedMSFileLoader

from glycresoft.structure.enums import AnalysisTypeEnum, GlycopeptideSearchStrategy

if TYPE_CHECKING:
    from glycresoft.composition_distribution_model.site_model.glycoproteome_model import GlycoproteomeModel


DillType = partial(PickleType, pickler=dill)


class _AnalysisParametersProps:
    def open_ms_file(self) -> Optional[ProcessedRandomAccessScanSource]:
        path = self.parameters.get("sample_path")
        if not path:
            return None
        if os.path.exists(path):
            return ProcessedMSFileLoader(path)
        return None

    def glycosite_model_path(self) -> Optional['GlycoproteomeModel']:
        from glycresoft.composition_distribution_model.site_model.glycoproteome_model import (
            GlycoproteomeModel, GlycosylationSiteModel)

        path = self.parameters.get('glycosylation_site_models_path')
        if os.path.exists(path):
            with open(path, 'rt', encoding='utf8') as fh:
                site_models = GlycosylationSiteModel.load(fh)
            target_model = GlycoproteomeModel.bind_to_hypothesis(
                object_session(self),
                site_models,
                hypothesis_id=self.hypothesis_id,
                fuzzy=True)
            return target_model
        return None

    @property
    def mass_shifts(self):
        return self.parameters.get('mass_shifts', [])

    @property
    def ms1_scoring_model(self):
        return self.parameters.get('scoring_model')

    @property
    def msn_scoring_model(self):
        return self.parameters.get('tandem_scoring_model')

    @property
    def retention_time_model(self):
        return self.parameters.get('retention_time_model')

    @property
    def fdr_estimator(self):
        return self.parameters.get('fdr_estimator')

    @property
    def search_strategy(self):
        strategy_str = self.parameters.get('search_strategy')
        strategy = GlycopeptideSearchStrategy[strategy_str]
        return strategy

    @property
    def is_multiscore(self):
        strategy = self.search_strategy
        return strategy == GlycopeptideSearchStrategy.multipart_target_decoy_competition


class Analysis(Base, HasUniqueName, HasFiles, HasParameters, _AnalysisParametersProps):
    __tablename__ = "Analysis"

    id: int = Column(Integer, primary_key=True)
    sample_run_id: int = Column(Integer, ForeignKey(SampleRun.id, ondelete="CASCADE"), index=True)
    sample_run: SampleRun = relationship(SampleRun)
    analysis_type: str = Column(String(128))
    status: str = Column(String(28))

    def __init__(self, **kwargs):
        self._init_parameters(kwargs)
        super(Analysis, self).__init__(**kwargs)

    def __repr__(self):
        sample_run = self.sample_run
        if sample_run:
            sample_run_name = sample_run.name
        else:
            sample_run_name = "<Detached From Sample>"
        return "Analysis(%s, %s)" % (self.name, sample_run_name)

    def _infer_hypothesis_id(self):
        try:
            hypothesis_id = self.parameters['hypothesis_id']
            return hypothesis_id
        except KeyError:
            session = object_session(self)
            if self.analysis_type == AnalysisTypeEnum.glycopeptide_lc_msms:
                from . import IdentifiedGlycopeptide, GlycopeptideHypothesis, Glycopeptide
                hypothesis_id = session.query(func.distinct(GlycopeptideHypothesis.id)).join(
                    Glycopeptide).join(
                    IdentifiedGlycopeptide,
                    Glycopeptide.id == IdentifiedGlycopeptide.structure_id).filter(
                    IdentifiedGlycopeptide.analysis_id == self.id).scalar()
                return hypothesis_id
            elif self.analysis_type == AnalysisTypeEnum.glycan_lc_ms:
                from . import GlycanComposition, GlycanCompositionChromatogram, GlycanHypothesis
                hypothesis_id = session.query(func.distinct(GlycanHypothesis.id)).join(GlycanComposition).join(
                    GlycanCompositionChromatogram,
                    GlycanCompositionChromatogram.glycan_composition_id == GlycanComposition.id).filter(
                    GlycanCompositionChromatogram.analysis_id == self.id).scalar()
                return hypothesis_id
            else:
                raise ValueError(self.analysis_type)

    @property
    def hypothesis_id(self):
        return self._infer_hypothesis_id()

    def get_hypothesis(self):
        from . import GlycopeptideHypothesis, GlycanHypothesis
        hypothesis_id = self.hypothesis_id
        if hypothesis_id is None:
            raise ValueError(
                "Analysis does not have a Hypothesis (analysis id: %r, name: %r)" % (
                    self.id, self.name))
        session = object_session(self)
        if self.analysis_type == AnalysisTypeEnum.glycopeptide_lc_msms:
            return session.query(GlycopeptideHypothesis).get(hypothesis_id)
        elif self.analysis_type == AnalysisTypeEnum.glycan_lc_ms:
            return session.query(GlycanHypothesis).get(hypothesis_id)

    hypothesis = property(get_hypothesis)

    def aggregate_identified_glycoproteins(self, glycopeptides=None):
        from glycresoft.tandem.glycopeptide.identified_structure import IdentifiedGlycoprotein
        from . import Protein
        if glycopeptides is None:
            glycopeptides = self.identified_glycopeptides.all()

        protein_index = {}
        session = object_session(self)
        proteins = session.query(Protein).all()
        for p in proteins:
            protein_index[p.id] = p

        glycoproteome = IdentifiedGlycoprotein.aggregate(glycopeptides, index=protein_index)
        return glycoproteome



class BoundToAnalysis(object):

    @declared_attr
    def analysis_id(self):
        return Column(Integer, ForeignKey(Analysis.id, ondelete="CASCADE"), index=True)

    @declared_attr
    def analysis(self):
        return relationship(Analysis, backref=backref(self._collection_name(), lazy='dynamic'))

    @classmethod
    def _collection_name(cls):
        name = cls.__name__
        collection_name = re.sub(r"(.+?)([A-Z])", lambda match: match.group(1).lower() +
               "_" + match.group(2).lower(), name, 0) + 's'
        return collection_name



from typing import Any, List, Optional, Sequence, Set, Tuple, Union, Generic, TypeVar, DefaultDict, Iterable

from collections import defaultdict

import numpy as np

import ms_deisotope
from glycopeptidepy.structure.fragment import FragmentBase


K = TypeVar("K")
V = TypeVar("V")



class PeakFragmentPair(object):
    __slots__ = ["peak", "fragment", "fragment_name", "_hash"]

    peak: ms_deisotope.DeconvolutedPeak
    fragment: Union[FragmentBase, Any]
    fragment_name: str
    _hash: int

    def __init__(self, peak, fragment):
        self.peak = peak
        self.fragment = fragment
        self.fragment_name = fragment.name
        self._hash = hash(self.peak)

    def __eq__(self, other):
        return (self.peak == other.peak) and (self.fragment_name == other.fragment_name)

    def __hash__(self):
        return self._hash

    def __reduce__(self):
        return self.__class__, (self.peak, self.fragment)

    def clone(self):
        return self.__class__(self.peak, self.fragment)

    def __repr__(self):
        return "PeakFragmentPair(%r, %r)" % (self.peak, self.fragment)

    def __iter__(self):
        yield self.peak
        yield self.fragment

    def mass_accuracy(self):
        return (self.peak.neutral_mass - self.fragment.mass) / self.fragment.mass


class _FragmentIndexBase(Generic[K, V]):
    _mapping: DefaultDict[K, List[V]]
    fragment_set: Iterable[PeakFragmentPair]

    def __init__(self, fragment_set):
        self.fragment_set = fragment_set
        self._mapping = None

    @property
    def mapping(self):
        if self._mapping is None:
            self._create_mapping()
        return self._mapping

    def invalidate(self):
        self._mapping = None

    def __getitem__(self, key):
        return self.mapping[key]

    def __iter__(self):
        return iter(self.mapping)

    def items(self):
        return self.mapping.items()

    def keys(self):
        return self.mapping.keys()

    def values(self):
        return self.mapping.values()

    def __len__(self):
        return len(self.mapping)

    def __contains__(self, key):
        return key in self.mapping

    def __str__(self):
        return str(self.mapping)


class ByFragmentIndex(_FragmentIndexBase[Union[FragmentBase, Any], ms_deisotope.DeconvolutedPeak]):

    def _create_mapping(self):
        self._mapping = defaultdict(list)
        for peak_fragment_pair in self.fragment_set:
            self._mapping[peak_fragment_pair.fragment].append(
                peak_fragment_pair.peak)


class ByPeakIndex(_FragmentIndexBase[ms_deisotope.DeconvolutedPeak, Union[FragmentBase, Any]]):

    def _create_mapping(self):
        self._mapping = defaultdict(list)
        for peak_fragment_pair in self.fragment_set:
            self._mapping[peak_fragment_pair.peak].append(
                peak_fragment_pair.fragment)


class FragmentMatchMap(object):
    members: Set[PeakFragmentPair]
    by_fragment: ByFragmentIndex
    by_peak: ByPeakIndex

    def __init__(self):
        self.members = set()
        self.by_fragment = ByFragmentIndex(self)
        self.by_peak = ByPeakIndex(self)

    def add(self, peak: Union[ms_deisotope.DeconvolutedPeak, PeakFragmentPair],
            fragment: Optional[Union[FragmentBase, Any]]=None):
        if fragment is not None:
            peak = PeakFragmentPair(peak, fragment)
        if peak not in self.members:
            self.members.add(peak)
            self.by_peak.invalidate()
            self.by_fragment.invalidate()

    def pairs_by_name(self, name: str) -> List[PeakFragmentPair]:
        pairs = []
        for pair in self:
            if pair.fragment_name == name:
                pairs.append(pair)
        return pairs

    def fragments_for(self, peak):
        return self.by_peak[peak]

    def peaks_for(self, fragment):
        return self.by_fragment[fragment]

    def __eq__(self, other):
        return self.members == other.members

    def __ne__(self, other):
        return self.members != other.members

    def __iter__(self):
        return iter(self.members)

    def __len__(self):
        return len(self.members)

    def items(self):
        for peak, fragment in self.members:
            yield fragment, peak

    def values(self):
        for pair in self.members:
            yield pair.peak

    def fragments(self):
        fragments = set()
        for peak_pair in self:
            fragments.add(peak_pair.fragment)
        return fragments

    def remove_fragment(self, fragment):
        peaks = self.peaks_for(fragment)
        for peak in peaks:
            self.members.remove(PeakFragmentPair(peak, fragment))
        self.by_fragment.invalidate()
        self.by_peak.invalidate()

    def remove_peak(self, peak):
        fragments = self.fragments_for(peak)
        for fragment in fragments:
            self.members.remove(PeakFragmentPair(peak, fragment))
        self.by_peak.invalidate()
        self.by_fragment.invalidate()

    def copy(self):
        inst = self.__class__()
        for case in self.members:
            inst.add(case)
        return inst

    def clone(self):
        return self.copy()

    def __repr__(self):
        return "FragmentMatchMap(%s)" % (', '.join(
            f.name for f in self.fragments()),)

    def clear(self):
        self.members.clear()
        self.by_fragment.invalidate()
        self.by_peak.invalidate()


def count_peaks_shared(reference: FragmentMatchMap, alternate: FragmentMatchMap) -> Tuple[Set[int], Set[int], Set[int]]:
    ref_peaks = {p.index.neutral_mass for p in reference.by_peak.keys()}
    alt_peaks = {p.index.neutral_mass for p in alternate.by_peak.keys()}

    shared = ref_peaks & alt_peaks

    ref_peaks_ = ref_peaks - alt_peaks
    alt_peaks_ = alt_peaks - ref_peaks

    return shared, ref_peaks_, alt_peaks_


def intensity_from_peak_indices(spectrum: Sequence[ms_deisotope.DeconvolutedPeak], indices: Iterable[int]) -> float:
    return sum([spectrum[i].intensity for i in indices])


class PeakPairTransition(object):
    start: ms_deisotope.DeconvolutedPeak
    end: ms_deisotope.DeconvolutedPeak
    annotation: Any
    key: Tuple[int, int]

    _hash: int

    def __init__(self, start, end, annotation):
        self.start = start
        self.end = end
        self.annotation = annotation
        # The indices of the start peak and end peak
        self.key = (self.start.index.neutral_mass, self.end.index.neutral_mass)
        self._hash = hash(self.key)

    def __eq__(self, other):
        if self.key != other.key:
            return False
        elif self.start != other.start:
            return False
        elif self.end != other.end:
            return False
        elif self.annotation != other.annotation:
            return False
        return True

    def __iter__(self):
        yield self.start
        yield self.end
        yield self.annotation

    def __hash__(self):
        return self._hash

    def __repr__(self):
        return ("{self.__class__.__name__}({self.start.neutral_mass:0.3f} "
                "-{self.annotation}-> {self.end.neutral_mass:0.3f})").format(self=self)


class SpectrumGraph(object):
    def __init__(self):
        self.transitions = set()
        self.by_first = defaultdict(list)
        self.by_second = defaultdict(list)

    def add(self, p1, p2, annotation):
        p1, p2 = sorted((p1, p2), key=lambda x: x.index.neutral_mass)
        trans = PeakPairTransition(p1, p2, annotation)
        self.transitions.add(trans)
        self.by_first[trans.key[0]].append(trans)
        self.by_second[trans.key[1]].append(trans)

    def __iter__(self):
        return iter(self.transitions)

    def __len__(self):
        return len(self.transitions)

    def __repr__(self):
        return "{self.__class__.__name__}({size})".format(
            self=self, size=len(self.transitions))

    def _get_maximum_index(self):
        try:
            trans = max(self.transitions, key=lambda x: x.key[1])
            return trans.key[1]
        except ValueError:
            return 0

    def adjacency_matrix(self):
        n = self._get_maximum_index() + 1
        A = np.zeros((n, n))
        for trans in self.transitions:
            A[trans.key] = 1
        return A

    def topological_sort(self, adjacency_matrix=None):
        if adjacency_matrix is None:
            adjacency_matrix = self.adjacency_matrix()
        else:
            adjacency_matrix = adjacency_matrix.copy()

        waiting = set()
        for i in range(adjacency_matrix.shape[0]):
            # Check for incoming edges. If no incoming
            # edges, add to the waiting set
            if adjacency_matrix[:, i].sum() == 0:
                waiting.add(i)
        ordered = list()
        while waiting:
            ix = waiting.pop()
            ordered.append(ix)
            outgoing_edges = adjacency_matrix[ix, :]
            indices_of_neighbors = np.nonzero(outgoing_edges)[0]
            # For each outgoing edge
            for neighbor_ix in indices_of_neighbors:
                # Remove the edge
                adjacency_matrix[ix, neighbor_ix] = 0
                # Test for incoming edges
                if (adjacency_matrix[:, neighbor_ix] == 0).all():
                    waiting.add(neighbor_ix)
        if adjacency_matrix.sum() > 0:
            raise ValueError("%d edges left over" % (adjacency_matrix.sum(),))
        else:
            return ordered

    def path_lengths(self):
        adjacency_matrix = self.adjacency_matrix()
        distances = np.zeros(self._get_maximum_index() + 1)
        for ix in self.topological_sort():
            incoming_edges = np.nonzero(adjacency_matrix[:, ix])[0]
            if incoming_edges.shape[0] == 0:
                distances[ix] = 0
            else:
                longest_path_so_far = distances[incoming_edges].max()
                distances[ix] = longest_path_so_far + 1
        return distances

    def paths_starting_at(self, ix):
        paths = []
        for trans in self.by_first[ix]:
            paths.append([trans])
        finished_paths = []
        while True:
            extended_paths = []
            for path in paths:
                terminal = path[-1]
                edges = self.by_first[terminal.key[1]]
                if not edges:
                    finished_paths.append(path)
                for trans in edges:
                    extended_paths.append(path + [trans])
            paths = extended_paths
            if len(paths) == 0:
                break
        return self.transitive_closure(finished_paths)

    def paths_ending_at(self, ix):
        paths = []
        for trans in self.by_second[ix]:
            paths.append([trans])
        finished_paths = []
        while True:
            extended_paths = []
            for path in paths:
                terminal = path[0]
                edges = self.by_second[terminal.key[0]]
                if not edges:
                    finished_paths.append(path)
                for trans in edges:
                    extended_paths.append([trans] + path)
            paths = extended_paths
            if len(paths) == 0:
                break
        return self.transitive_closure(finished_paths)

    def transitive_closure(self, paths):
        # precompute node_sets for each path
        node_sets = {}
        for i, path in enumerate(paths):
            # track all nodes by index in this path in node_set
            node_set = set()
            for node in path:
                # add the start node index and end node index to the
                # set of nodes on this path
                node_set.update(node.key)
            node_sets[i] = (node_set, len(path))
        keep = []
        node_sets_items = list(node_sets.items())
        for i, path in enumerate(paths):
            node_set, length = node_sets[i]
            is_enclosed = False
            for key, value_pair in node_sets_items:
                # attempting to be clever here seems to cost more
                # time than it saves.
                if key == i:
                    continue
                other_node_set, other_length = value_pair
                if node_set < other_node_set:
                    is_enclosed = True
                    break
            if not is_enclosed:
                keep.append(path)
        return sorted(keep, key=len, reverse=True)

    def longest_paths(self, limit=-1):
        # get all distinct paths
        paths = []
        for ix in np.argsort(self.path_lengths())[::-1]:
            segment = self.paths_ending_at(ix)
            paths.extend(segment)
            # remove redundant paths
            paths = self.transitive_closure(paths)
            if limit > 0:
                if len(paths) > limit:
                    break
            elif len(segment) == 0:
                break
        return paths


try:
    has_c = True
    _PeakFragmentPair = PeakFragmentPair
    _PeakPairTransition = PeakPairTransition
    _SpectrumGraph = SpectrumGraph
    _FragmentMatchMap = FragmentMatchMap

    from glycresoft._c.structure.fragment_match_map import (
        PeakFragmentPair, PeakPairTransition, SpectrumGraph, FragmentMatchMap)
except ImportError:
    has_c = False
